#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
from __future__ import division
import calendar
import cairo
from rayon import (
    backend_cairo,
    backgrounds,
    borders,
    charts,
    data,
    grids,
    lines,
    plots,
    scales,
    text,
    tickmarkers,
    ticks,
)
import netsa.data.times
import netsa.data.format

# TODO: modify prism so this isn't necessary
#text.use_old_get_point_size()


def get_context(fobj, output_type, width, height):
    # Figure out which backend we're working with
    if output_type == "svg":
        surface = cairo.SVGSurface(fobj, width, height)
        raster_output = False
    elif output_type == "pdf":
        surface = cairo.PDFSurface(fobj, width, height)
        raster_output = False
    elif output_type == "ps":
        surface = cairo.PSSurface(fobj, width, height)
        surface.set_size(width, height)
        raster_output = False
    elif output_type == "png":
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
        raster_output = True
    else:
        raise RuntimeError("Unknown image format for '%s'" % output_type)

    return surface, backend_cairo.Context(cairo.Context(surface),
                                          width, height, raster_output)





def get_subplot(all_data, x_scale,
                major_tick_data, minor_tick_data, title, left_border_width=0,
                padding=(0, 0, 0, 0),
                bgcolor=(0, 0, 0, 0), top_color=(255, 0, 0, 128),
                bottom_color=(255, 0, 0, 128), annotation_function=None,
                top_label="OUT", bottom_label="IN"):

    chart = charts.SquareChart(backend_cairo)

    # Columns
    col_time = all_data.column('time')
    col_top  = all_data.column('top')
    col_btm  = all_data.column('bottom')


    # Plots
    top_plot = plots.FilledLinePlot(top_color)
    btm_plot = plots.FilledLinePlot(bottom_color)
    ann_top_plot = plots.LabeledScatterPlot()
    ann_top_plot.set_options(avoid_horizontal_overflow=True)
    ann_btm_plot = plots.LabeledScatterPlot()
    ann_btm_plot.set_options(avoid_horizontal_overflow=True)


    # Limits and scales

    # X shares the time scale between both plots. Y should share a
    # single scale, limited by the largest point on either Y scale.
    y_hi = max(col_top.max(), col_btm.max())


    # Scales
    ytop_scale = scales.LinearScale(0, y_hi, .5, 1)
    ybtm_scale = scales.LinearScale(0, y_hi, .5, 0)

    # Set scales and data
    top_plot.set_scales(x=x_scale, y=ytop_scale,
                        line_color=(0, 0, 0), line_width=1,
                        line_style="solid")
    btm_plot.set_scales(x=x_scale, y=ybtm_scale,
                        line_color=(0, 0, 0), line_width=1,
                        line_style="solid")

    top_plot.set_data(x=col_time, y=col_top)
    btm_plot.set_data(x=col_time, y=col_btm)

    # Synthesize the dataset for the annotations
    if annotation_function is None:
        def annotation_function(row):
            return "%s, %d:%02d" % (
                netsa.data.format.num_prefix(row[1], "b/s"),
                row[0].hour, row[0].minute)

    ann_top_iter = ([r[0], r[1]] for r in all_data if r.top == col_top.max())
    ann_top_data = data.Dataset.from_rows([ann_top_iter.next()])
    col_ann_top_text = data.Column([annotation_function(row)
                                    for row in ann_top_data])

    ann_btm_iter = ([r[0], r[2]] for r in all_data if r.bottom == col_btm.max())
    ann_btm_data = data.Dataset.from_rows([ann_btm_iter.next()])
    col_ann_btm_text = data.Column([annotation_function(row)
                                    for row in ann_btm_data])

    ann_top_plot.set_scales(x=x_scale, y=ytop_scale,
                            marker_color=(0, 0, 0), marker_shape="circle",
                            marker_size=5, text=lambda s: s,
                            label_position="ne", label_spacing=5,
                            font_properties={'halign': 'left',
                                             'valign': 'center'},
                            # bgcolor=bgcolor,
                            border_properties={})
    ann_btm_plot.set_scales(x=x_scale, y=ybtm_scale,
                            marker_color=(0, 0, 0), marker_shape="circle",
                            marker_size=5, text=lambda s: s,
                            label_position="se", label_spacing=5,
                            font_properties={'halign': 'left',
                                             'valign': 'center'},
                            # bgcolor=bgcolor,
                            border_properties={})

    ann_top_plot.set_data(x=ann_top_data.column(0), y=ann_top_data.column(1),
                          text=col_ann_top_text)
    ann_btm_plot.set_data(x=ann_btm_data.column(0), y=ann_btm_data.column(1),
                          text=col_ann_btm_text)
    
                         

    chart.add_plot(top_plot, "top")
    chart.add_plot(btm_plot, "bottom")
    chart.add_plot(ann_top_plot, "ann_top")
    chart.add_plot(ann_btm_plot, "ann_btm")

    

    # Borders

    # In and Out labels
    inout_labels = (borders.create('vlabel', label=top_label, halign="center",
                                   valign="bottom"),
                    borders.create('vlabel', label=bottom_label,
                                   halign="center", valign="bottom"))
    yinout = borders.create('vsplit', children=inout_labels, rpad="1px")

    # One border on the left with a vertically centered,
    # right-justified title.
    yborder = borders.create(
        'vline', line=lines.create("solid", color=(0, 0, 0), width=2))
    ytitle = borders.create(
        'hlabel', label=title,
        font_weight="bold", font_size="large",
        halign="right", valign="center")
    chart.add_left_border(yborder, width=.001)
    chart.add_left_border(yinout, width=.015)
    chart.add_left_border(ytitle, width=left_border_width)

    # Add padding as necessary
    chart.add_custom_padding(*padding)


    # Tick marks and gridlines (Gridline is in the center of the plot,
    # not the border)
    minor_tickmarker = tickmarkers.SimpleHorizontalTickMarker(
        marker_args=dict(below=2, above=2))
    minor_tickset = ticks.Ticks([(d, "") for d in minor_tick_data],
                                x_scale, minor_tickmarker)

    major_tickmarker = tickmarkers.SimpleHorizontalTickMarker(
        marker_args=dict(below=4, above=4))
    major_tickset = ticks.Ticks([(d, "") for d in major_tick_data],
                                x_scale, major_tickmarker)

    centerline = grids.HorizontalGridLines()
    centerline.set_data(pos=data.Column([0]),
                        ticksets=data.Column([[minor_tickset,
                                               major_tickset]]))
    centerline.set_scales(pos=ytop_scale,
                          ticksets=lambda x: x,
                          line_style='solid',
                          line_width=1,
                          line_color=(0, 0, 0))

    daily_gridlines = grids.VerticalGridLines()
    daily_gridlines.set_data(pos=data.Column(major_tick_data))
    daily_gridlines.set_scales(pos=x_scale,
                               ticksets=lambda x: None,
                               line_style='dotted',
                               line_width=1,
                               line_color=(32, 32, 32))
                          
    chart.add_front_gridlines(centerline)
    chart.add_rear_gridlines(daily_gridlines)

    chart.set_chart_background(backgrounds.ColorBackground(bgcolor))

            
    return chart


def get_x_data(binned_time):
    # Ticks on X every four hours with a major tick every day.
    def as_seconds(d):
        return calendar.timegm(d.timetuple())
    col_time = data.Column(binned_time)
    x_lo, x_hi = col_time.min(), col_time.max()
    x_scale = scales.TimeScale(x_lo, x_hi)
    time_tick_data = [netsa.data.times.make_datetime(secs)
                      for secs in xrange(int(as_seconds(x_lo)),
                                         int(as_seconds(x_hi)),
                                         60 * 60 * 4)]
    minor_tick_data = [d  for d in time_tick_data
                       if (d.minute == 0 and d.second == 0 and
                           d.hour != 0 and d.hour % 4 == 0)]
                      
    major_tick_data = [d for d in time_tick_data
                        if (d.minute == 0 and d.second == 0 and
                            d.hour == 0)]
    # This tickmarker will be used by the stripplot_* functions to
    # draw a ticked border at the bottom of the chart XXX TODO FIXME:
    # By all rights, that should be valign="bottom". There's probably
    # a bug in either the vertical text alignment code in text.py
    # (get_bbox) or in where the TickMarker code chooses to place the
    # label.
    major_tickmarker = tickmarkers.SimpleHorizontalTickMarker(
        marker_args=dict(below=4, above=4, width=2), valign="top")
    major_tickset = ticks.Ticks([(d, "%02d/%02d" % (d.month, d.day))
                                 for d in major_tick_data],
                                x_scale, major_tickmarker)
    return col_time, x_scale, major_tickset, major_tick_data, minor_tick_data
                                             



def stripplot_single(binned_time, top, bottom, title,
                     fobj, output_type, width, height,
                     padding=(0, 0, 0, 0),
                     title_width=0,
                     bgcolor=(255,255,255),
                     top_color=(255, 0, 0, 127),
                     bottom_color=(0, 0, 255, 127),
                     annotation_function=None):
    """
    Write a single stripplot to a stream.

    @type  binned_time: list of L{datetime.datetime}
    @param binned_time: sequence of times to plot on the X axis

    @type  top: list of numbers, or L{data.Column}
    @param top: sequence of numbers to plot on the upper Y axis

    @type  bottom: list of numbers, or L{data.Column}
    @param bottom: sequence of numbers to plot on the lower Y axis

    @type  title: string
    @param title: title to print on the left side of the plot

    @type  fobj: file-like object
    @param fobj: open file into which to output the image data

    @type  output_type: string literal: C{svg}, C{pdf} or C{png}
    @param output_type: output format of image data

    @type  width: int
    @param width: width of output image

    @type  height: int
    @param height: height of output image

    @type  padding: 4-tuple
    @param padding: Additional left, right, top and bottom padding for
        the chart, in device units (points or pixels).

    @type  title_width: int
    @param title_width: width of left border, which contains the
        title. Given in device units (points or pixels).

    @type  bgcolor: RGB or RGBA tuple
    @param bgcolor: color of the background

    @type  top_color: RGB or RGBA tuple
    @param top_color: color of the upper data plot

    @type  bottom_color: RGB or RGBA tuple
    @param bottom_color: color of the lower data plot

    @type  annotation_function: function of L{Dataset.Row} -> string
    @param annotation_function: function that takes a data point and
        returns a text annotation of that data point. If this is
        C{None} (the default), a default function is used.
    """
    page = backend_cairo.Page.from_file_object(
        fobj, output_type, width, height)

    # Get proportions from pixels widths
    lpad, rpad = [p / width for p in padding[:2]]
    tpad, bpad = [p / height for p in padding[2:]]
    title_width = title_width / width

    (col_time,
     x_scale,
     major_tickset,
     major_tick_data,
     minor_tick_data) = get_x_data(binned_time)

    if type(top) == data.Column:
        col_top = top
    else:
        col_top = data.Column(top)
    if type(bottom) == data.Column:
        col_btm = bottom
    else:
        col_btm = data.Column(bottom)
    all_data = data.Dataset([col_time, col_top, col_btm],
                            header=dict(title=title),
                            colnames=['time', 'top', 'bottom'])
            


    chart = get_subplot(all_data, x_scale,
                        major_tick_data, minor_tick_data, title, title_width,
                        (lpad, rpad, tpad, bpad),
                        bgcolor, top_color, bottom_color, annotation_function)
    
    bottom_border = borders.create("hline",
                                   line=lines.create("solid", width=2))
    bottom_border.add_tickset(major_tickset)
    
    chart.add_bottom_border(bottom_border, height="20px")

    page.write(chart)
        

def stripplot_multi(binned_time, tops, bottoms, titles,
                    fobj, output_type, width, height,
                    vertical_padding=0, horizontal_padding=0,
                    title_width=0,
                    bgcolor=(255, 255, 255), subplot_bgcolor=None,
                    top_color=(255, 0, 0, 127), bottom_color=(0, 0, 255, 127),
                    annotation_function=None, chart_title=""):
    """
    Write multiple stripplots to a stream.

    @type  binned_time: list of L{datetime.datetime}
    @param binned_time: sequence of times to plot on the X axis

    @type  top: list of lists of numbers, or list of L{data.Column}
    @param top: sequence of sequences of numbers to plot on the upper
        Y axis Each sequence represents input for a single subplot

    @type  bottom: list of lists of numbers, or list of L{data.Column}
    @param bottom: sequence of sequences of numbers to plot on the
        lower Y axis Each sequence represents input for a single
        subplot

    @type  titles: list of strings
    @param titles: titles to print on the left side of the plot
        Each title represents input for a single subplot

    @type  fobj: file-like object
    @param fobj: open file into which to output the image data

    @type  output_type: string literal: C{svg}, C{pdf}, C{ps} or C{png}
    @param output_type: output format of image data

    @type  width: int
    @param width: width of output image in device units (points or
        pixels)

    @type  height: int
    @param height: height of output image in device units (points or
        pixels)

    @type  vertical_padding: int
    @param vertical_padding: vertical padding between subplots, in
        device units (points or pixels).

    @type  horizontal_padding: int
    @param horizontal_padding: horizontal padding to the left and
        right of the stack of plots, in device units (points or pixels).

    @type  title_width: int
    @param title_width: left border width, reserved for the title.
        This value will be used for all the plots.

    @type  bgcolor: RGB or RGBA tuple
    @param bgcolor: color of the background

    @type  subplot_bgcolor: RGB or RGBA tuple
    @param subplot_bgcolor: color of the background of the individual
        subplots. If not specified, the value of C{bgcolor} will be used.

    @type  top_color: RGB or RGBA tuple
    @param top_color: color of the upper data plot. This value will be
        used for all the plots.

    @type  bottom_color: RGB or RGBA tuple
    @param bottom_color: color of the lower data plot. This value will
        be used for all the plots.

    @type  annotation_function: function of L{Dataset.Row} -> string
    @param annotation_function: function that takes a data point and
        returns a text annotation of that data point. If this is
        C{None} (the default), a default function is used.

    @type  chart_title: string
    @param chart_title: Title of the overall visualization
    """
    if (len(tops) != len(bottoms) or len(titles) != len(bottoms)):
        raise RuntimeError("different lengths for titles (%d), tops (%d) "
                           "and bottoms (%d)" % (
            len(titles), len(tops), len(bottoms)))

    # Normalize subplot_bgcolor
    if subplot_bgcolor is None:
        subplot_bgcolor = bgcolor
    
    page = backend_cairo.Page.from_file_object(
        fobj, output_type, width, height)
    

    vpad = (vertical_padding * len(tops)) / height
    hpad = horizontal_padding / width

    multichart = charts.TiledChart(backend=backend_cairo,
                                   col_weights=[1], row_weights=[],
                                   orientation="vertical")
                                   

    title_width = title_width / width
    lpad = rpad = horizontal_padding / width / 2
    tpad = bpad = vertical_padding / height / 2

    # Merge vpad into tpad and bpad
    tpad += vpad / 2
    bpad += vpad / 2

    # Precalculate time ticks
    (col_time,
     x_scale,
     major_tickset,
     major_tick_data,
     minor_tick_data) = get_x_data(binned_time)

    for top, bottom, title in zip(tops, bottoms, titles):
        if type(top) == data.Column:
            col_top = top
        else:
            col_top = data.Column(top)
        if type(bottom) == data.Column:
            col_btm = bottom
        else:
            col_btm = data.Column(bottom)
        all_data = data.Dataset([col_time, col_top, col_btm],
                                header=dict(title=title),
                                colnames=['time', 'top', 'bottom'])
        chart = (get_subplot(all_data, x_scale,
                             major_tick_data, minor_tick_data,
                             title, title_width,
                             (lpad, rpad, tpad, bpad),
                             subplot_bgcolor, top_color, bottom_color,
                             annotation_function))
        multichart.add_chart(chart)


    # Add a white background to the output
    multichart.set_chart_background(backgrounds.ColorBackground(bgcolor))

    # Add a title
    if chart_title != "":
        multichart.add_top_border(
            borders.create('hlabel', label=chart_title, font_size="large",
                           valign="top"),
            height=.05)

    # Cheap hack -- the last chart is left over in chart, so put a
    # bottom border on it.
    bottom_border = borders.create("hline",
                                   line=lines.create("solid", width=2))
    bottom_border.add_tickset(major_tickset)
    
    chart.add_bottom_border(bottom_border, height="20px")


    page.write(multichart)




__all__ = ["stripplot_single", "stripplot_multi"]
