#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
from __future__ import division

import sys
import math
import os

try:
    import cairo
    def is_available(): return True
except ImportError:
    def is_available(): return False
    


from rayon import RayonException, backend, raster

# Every backend has a name
name = "cairo"

# Cairo supports static (file) output. (Indeed, that's all it does.)
# As a result, it has a Page object which writes the static data to
# the output source.
output_to_file = True

# Cairo does not support output to memory right now. Well, we don't
# support it, at any rate.
output_to_memory = False

# Freetype support and font finding through fontconfig)
import fonts

# The Cairo backend is non-interactive -- no events, etc.
interactive = False

# font hackery -- You can mostly ignore this stuff; it's pretty much
# useless right now.

# TODO: get value via ctypes?
FT_Err_Ok = 0
CAIRO_STATUS_SUCCESS = 0


font_families = {
    "sans_serif" : ["Bitstream Vera Sans", "Helvetica", "Arial"],
    "serif"      : ["Tahoma", "Times", "Times New Roman"],
    "cursive"    : ["Tahoma", "Times", "Times New Roman"],
    "fantasy"    : ["Tahoma", "Times", "Times New Roman"],
    "monospace"  : ["Tahoma", "Times", "Times New Roman"],
}
font_families['default'] = font_families['sans_serif'] + font_families['serif']

# TODO: Something with styles and weights.....

font_styles = {
    "normal" : "normal",
    "slant"  : "slant",
    "italic" : "italic"
}

font_weights = {
    "normal" : "normal",
    "light"  : "light",
    "bold"   : "bold"
}


def pathmunge(old_path, pathspec):
    paths = pathspec.split(":")
    if pathspec.startswith(":"):
        return old_path + paths[1:]
    elif pathspec.endswith(":"):
        return paths[:-1] + paths
    else:
        return paths


if sys.platform == 'darwin':
    font_search_path = [
        "/Library/Fonts",
        "/System/Library/Fonts",
    ]
elif sys.platform == 'win32':
    font_search_path = [
        "/usr/share/fonts",
        "/usr/local/share/fonts",
    ]
elif sys.platform == 'linux2':
    font_search_path = [
        "/usr/share/fonts",
        "/usr/local/share/fonts",
    ]



if os.getenv("FONT_PATH"):
    font_search_path = pathmunge(font_search_path, os.getenv("FONT_PATH"))

def find_font(family, style, weight):
    # TODO: Ignoring style and weight for now
    #print "family: %s" % family
    for font in font_families[family]:
        #print "font: %s" % font
        # Sorry, pfm, we only do TrueType
        for ext in (".ttf", ".otf"):
            name = font + ext
            for path in font_search_path:
                joined = os.path.join(path, name)
                #print "\t" + joined + "?...",
                if os.path.exists(joined):
                    #print "Match!"
                    return joined
                else:
                    #print "Nope"
                    pass
    raise RuntimeError("Couldn't locate font for family '%s'" % family)




#
# Backend configuration
#


def half_round(x):
    # Round to nearest .5
    x = round(x * 2)
    if x % 2 == 0:
        x += 1
    return x / 2.0


def fixup_line_for_raster(x1, y1, x2, y2):
    if x1 == x2 or y1 == y2:
        #return tuple(half_round(_) for _ in (x1, y1, x2, y2))
        return tuple(round(_) for _ in (x1, y1, x2, y2))


#
# Cairo "event translator" (dummy)
#

# Deprecated
class EventTranslator(object):
    def get_name(self, evt): return ""
    def is_mouse_event(self, evt): return False

# New-style (These shouldn't ever be called really, calling code
# should wrap them in "if backend.is_interactive():..."
def get_event_name(evt): return ""
def is_mouse_event(evt): return False

#
# Cairo "tooltips" (dummy)
#

def tooltip(tooltippable, message):
    # What are these "tooltips" of which you speak?
    pass

#
# Cairo Context
#

class Context(object):
    """An abstraction over the cairo.Context. """
    def __init__(self, cairo_context, abswidth, absheight, raster_output=False):
        self.ctx = cairo_context
        self.abswidth = abswidth
        self.absheight = absheight
        self.interactive = interactive
        self.raster_output = raster_output
        self.dashes = None
    def canvas_dimensions(self):
        return (self.abswidth, self.absheight)
    def normalize_dimensions(self, width, height):
        if width is None:
            width = self.abswidth
        if height is None:
            height = self.absheight
        return width, height
    # Transformation matrix stuff
    def translate(self, dx, dy, fixup=False):
        if self.raster_output and fixup:
            dx, dy = tuple(half_round(_) for _ in (dx, dy))
        return self.ctx.translate(dx, dy)
    def scale(self, dx, dy):
        return self.ctx.scale(dx, dy)
    # Resource stacks
    def push_state(self):
        return self.ctx.save()
    def pop_state(self):
        return self.ctx.restore()
    def _init_dash(self):
        # Do nothing if already initialized
        if self.dashes is not None:
            return
        # Maintain the current dash at the top of the stack.  (Older
        # versions of pycairo don't have get_dash, so we can't
        # maintain the top of the dash stack in the context.)
        self.dashes = []
        try:
            self.dashes.append(self.ctx.get_dash())
        except AttributeError:
            self.dashes.append(((), 0.0))
    def _push_dash(self, new_dash):
        self._init_dash()
        self.dashes.append(new_dash)
        self.ctx.set_dash(*new_dash)
    def _pop_dash(self):
        self._init_dash()
        popped_dash = self.dashes.pop(-1)
        curr_dash = self.dashes[-1]
        self.ctx.set_dash(*curr_dash)
        return popped_dash
    def _set_source_color(self, color):
        # Scale to 0..1
        color = [c/255.0 for c in color]
        if len(color) == 3:
            self.ctx.set_source_rgb(*color)
        else:
            self.ctx.set_source_rgba(*color)
        # Return color for debugging
        return color
    # Drawing proxies
    def draw_lines(self, segments, color, width, fixup=False,
                   dashes=None, dash_offset=0):
        #print "draw_lines(%s, %s, %s, fixup=%s, dashes=%s)" % (
        #    str(segments), color, width, fixup, dashes)
        self.ctx.save()
        if dashes is not None:
            self._push_dash((dashes, dash_offset))
        try:
            for x1, y1, x2, y2 in segments:
                # Clean up perfectly vertical and horizontal lines in
                # raster output
                if self.raster_output and fixup:
                    x1, y1, x2, y2 = fixup_line_for_raster(x1, y1, x2, y2)
                self.ctx.move_to(x1, y1)
                self.ctx.line_to(x2, y2)
            self._set_source_color(color)
            # trick to proof against possible uniform scale transforms
            self.ctx.set_line_width(
                max(self.ctx.device_to_user_distance(width, width)))
            self.ctx.stroke()
        finally:
            if dashes is not None:
                self._pop_dash()
            self.ctx.restore()

    def draw_shape(self,
                   segments,
                   start_point,
                   stroke_width,
                   stroke_color,
                   fill_color):
        self.ctx.save()
        try:
            self.ctx.move_to(*start_point)
            for action, args in segments:
                if action == "line":
                    x, y = args
                    self.ctx.line_to(x, y)
                elif action == "arc":
                    x, y, radius, angle_start, angle_end, clockwise = args
                    angle_start = 2*math.pi - math.radians(angle_start)
                    angle_end = 2*math.pi - math.radians(angle_end)
                    if not clockwise:
                        self.ctx.arc(x, y, radius, angle_start, angle_end)
                    else:
                        self.ctx.arc_negative(x,
                                              y,
                                              radius,
                                              angle_start,
                                              angle_end)
                # TODO: "curve"
                else:
                    raise RayonException("Unknown shape action: '%s'" % action)

            
            self._set_source_color(fill_color)
            self.ctx.fill_preserve()
            self._set_source_color(stroke_color)
            # TODO: May need to use max(device_to_user_distance(width,
            # width)) trick here
            self.ctx.set_line_width(stroke_width)
            self.ctx.stroke()
        finally:
            self.ctx.restore()
            
    def draw_polygon(self, vertices, stroke_width, stroke_color, fill_color,
                     fixup=False):
        self.ctx.save()
        if self.raster_output and fixup:
            vert_iter = (tuple(half_round(_) for _ in v) for v in vertices)
        else:
            vert_iter = (v for v in vertices)
        # Draw path
        try:
            try:
                first = vert_iter.next()
            except StopIteration:
                # No vertices
                pass
            else:
                #print "\tV: %s" % str(first)
                self.ctx.move_to(*first)
                for v in vert_iter:
                    #print "\tV: -> %s" % str(v)
                    self.ctx.line_to(*v)
                #print "\tV: => %s" % str(v)
                self.ctx.line_to(*first)
        finally:
            self.ctx.restore()
        # Fill and stroke path
        self.ctx.save()
        try:
            self._set_source_color(fill_color)
            self.ctx.fill_preserve()
            self._set_source_color(stroke_color)
            self.ctx.set_line_width(stroke_width)
            self.ctx.stroke()
        finally:
            self.ctx.restore()

    def draw_rectangle(self, x, y, width, height,
                       stroke_width, stroke_color, fill_color,
                       fixup=False):
        w = width
        h = height
        # (x, y) == "lower left" (If h > y, (x,y) will be upper left)
        vertices = ((x,y-h), (x,y), (x+w,y), (x+w,y-h))
        return self.draw_polygon(vertices,
                                 stroke_width, stroke_color, fill_color,
                                 fixup)
    def draw_line(self, x1, y1, x2, y2, color, width, fixup=False,
                  dashes=None, dash_offset=0):
        return self.draw_lines([[x1, y1, x2, y2]], color, width, fixup,
                               dashes, dash_offset)
    def draw_ellipse(self, x, y, width, height, stroke_color, fill_color,
                     stroke_width=1):
        w = width
        h = height
        # Create the path
        self.ctx.save()
        try:
            self.ctx.translate(x, y)
            self.ctx.scale(w/2.0, h/2.0)
            self.ctx.arc(0.0, 0.0, 1.0, 0.0, 2.0 * math.pi)
        finally:
            # Restore before doing the fill and stroke so they aren't
            # affected by the translate and scale
            self.ctx.restore()

        self.ctx.save()
        try:
            self._set_source_color(fill_color)
            self.ctx.fill_preserve()
            self._set_source_color(stroke_color)
            self.ctx.set_line_width(stroke_width)
            self.ctx.stroke()
        finally:
            self.ctx.restore()
    def font_families(self):
        return font_families.keys()
    def font_styles(self):
        return font_styles.keys()
    def font_weights(self):
        return font_weights.keys()
    def draw_text(self, text, bbox, size, family, style, weight,
                  color=(0, 0, 0), angle=None):
        self.ctx.save()
        try:
            x, y = bbox[1]
            # TODO: when you get around to making font switching work,
            # make sure this honors the same values backend_wx honors.
            #
            #fonts.switch_font(self.ctx, family, style, weight, size)
            self.ctx.set_font_size(size)
            _, _, _, h = self.ctx.text_extents(text)[:4]
            # Drawing text in cairo starts at the baseline, not the
            # top edge, as in wx.
            self.ctx.move_to(x, y)
            if angle:
                angle = 360 - angle
                self.ctx.rotate(math.radians(angle))
            self.ctx.text_path(text)
            self._set_source_color(color)
            self.ctx.fill()
            #self.ctx.show_text(text)
            # TODO: change the %$#@ing font back!
        finally:
            self.ctx.restore()
    def get_text_dimensions(self, text, size, family, style, weight):
        self.ctx.save()
        try:
            self.ctx.set_font_size(size)
            # Cairo extents are: (x_bearing, y_bearing, width, height,
            # ...some other stuff...)
            return self.ctx.text_extents(text)[2:4]
            # TODO: change the font back!
        finally:
            self.ctx.restore()
            
    def draw_bitmap(self, bitmap, x, y, width, height, fill):
        if bitmap.is_old():
            bitmap.set_format(format="rgba",
                              premultiply_alpha=True,
                              for_png=True)
            buff = bitmap.get_data()
        else:
            # TODO: Cairo ImageSurfaces store data in host byte
            # order. (IOW, "cairo.FORMAT_ARGB32" really means BGRA32
            # on little-endian systems.)
            if sys.byteorder == "little":
                buff = bitmap.get_data(fmt=raster.BGRA32_PM)
            else:
                buff = bitmap.get_data(fmt=raster.ARGB32_PM)
            
        image = cairo.ImageSurface.create_for_data(buff,
                                                   cairo.FORMAT_ARGB32,
                                                   bitmap.get_width(),
                                                   bitmap.get_height())
        image.flush()
        self.ctx.save()
        try:
            self.ctx.set_source_surface(image, x, y)
            if (fill == "scale" and
                (bitmap.get_width() != width or
                 bitmap.get_height() != height)):
                matrix = cairo.Matrix(xx=bitmap.get_width() / width,
                                      yy=bitmap.get_height() / height)
                self.ctx.get_source().set_matrix(matrix)
            self.ctx.paint()
        finally:
            self.ctx.restore()
            
        
    def draw_image(self, filename, x, y, width, height, fill):
        # Valid values for fill are: 'scale', 'tile', 'center' or
        # 'none'. I'm only implementing 'scale' and 'none' right now.
        image = cairo.ImageSurface.create_from_png(filename)
        self.ctx.set_source_surface(image, x, y)
        self.ctx.save()
        try:
            self.ctx.set_source_surface(image, x, y)
            if fill == "scale":
                matrix = cairo.Matrix(xx=image.get_width() / width,
                                      yy=image.get_height() / height)
                self.ctx.get_source().set_matrix(matrix)
            self.ctx.paint()
        finally:
            self.ctx.restore()
    # Cairo-specific functions
    def get_cairo_surface(self):
        return self.ctx.get_group_target()

#
# Convenience functions for dealing with Cairo
#

class FilenameMetaInfo(object):
    def __init__(self, fname):
        self.fname = fname
    def target(self): return self.fname
    def png(self): return self.fname.endswith(".png")
    def pdf(self): return self.fname.endswith(".pdf")
    def ps(self): return self.fname.endswith(".ps")
    def svg(self): return self.fname.endswith(".svg")
        

class FileObjectMetaInfo(object):
    def __init__(self, fobj, ext):
        self.fobj = fobj
        self.ext = ext
    def target(self): return self.fobj
    def png(self): return self.ext == "png"
    def pdf(self): return self.ext == "pdf"
    def ps(self): return self.ext == "ps"
    def svg(self): return self.ext == "svg"
    
    
def bootstrap_common(surface, width, height, raster_output):
    cairo_ctx = cairo.Context(surface)
    # This hack is necessary in some later versions of Cairo, which
    # seem to misallocate font resources (resulting in assertion
    # falures and SIGABRT) if you don't do this
    cairo_ctx.select_font_face('sans-serif')
    ctx = Context(cairo_ctx, width, height, raster_output)
    return ctx
    

def bootstrap_from_filename(fname, width, height):
    """Initialize the backend for drawing to a file."""
    if fname.endswith(".svg"):
        surface = cairo.SVGSurface(fname, width, height)
        raster_output = False
    elif fname.endswith(".pdf"):
        surface = cairo.PDFSurface(fname, width, height)
        raster_output = False
    elif fname.endswith(".png"):
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
        raster_output = True
    else:
        raise RayonException("Unknown image format for '%s'" % fname)
    
    ctx = bootstrap_common(surface, width, height, raster_output)
    ctx._meta = FilenameMetaInfo(fname)
    return ctx

def bootstrap_from_file_object(fobj, output_type, width, height):
    if output_type == "svg":
        surface = cairo.SVGSurface(fobj, width, height)
        raster_output = False
    elif output_type == "pdf":
        surface = cairo.PDFSurface(fobj, width, height)
        raster_output = False
    elif output_type == "ps":
        surface = cairo.PSSurface(fobj, width, height)
        surface.set_size(width, height)
        raster_output = False
    elif output_type == "png":
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
        raster_output = True
    else:
        raise RayonException("Unknown image format for '%s'" % output_type)

    ctx = bootstrap_common(surface, width, height, raster_output)
    ctx._meta = FileObjectMetaInfo(fobj, output_type)
    return ctx
    

def finalize(ctx):
    """Finalize the rendering of an image to a file, if necessary."""
    if hasattr(ctx, "_meta"):
        meta = ctx._meta
        if meta.png():
            ctx.get_cairo_surface().write_to_png(meta.target())
        if ((meta.pdf() or meta.ps() or meta.svg())
            and cairo.version_info < (1,3)):
            # _This_ hack is necessary because older versions of Cairo
            # seem to require this extra method call.
            ctx.ctx.show_page()
            ctx.get_cairo_surface().finish()


#
# Page object
#

class Page(backend.Page):
    @classmethod
    def from_filename(cls, filename, width, height):
        return FilePage(filename, width, height)
    @classmethod
    def from_file_object(cls, fobj, output_type, width, height):
        return FileObjectPage(fobj, output_type, width, height)

class FilePage(Page):
    def __init__(self, filename, width, height):
        self.filename = filename
        self.width = width
        self.height = height
    def write(self, chart):
        ctx = bootstrap_from_filename(
            self.filename, self.width, self.height)
        chart.layout(ctx=ctx, 
                     topleft=(0,0), 
                     width=None, 
                     height=None, 
                     call_depth=0)
        finalize(ctx)

class FileObjectPage(Page):
    def __init__(self, fobj, output_type, width, height):
        self.fobj = fobj
        self.output_type = output_type
        self.width = width
        self.height = height
    def write(self, chart):
        ctx = bootstrap_from_file_object(self.fobj, self.output_type,
                                         self.width, self.height)
        chart.layout(ctx,
                     topleft=(0,0),
                     width=None,
                     height=None,
                     call_depth=0)
        finalize(ctx)

        
#
# Events
#


