#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@

import plots
import lines
import scales
import borders

from rayon import RayonException

class GridLines(plots.Plot):
    """
    .. csv-table:: Axes
       :header: "Axis", "Value", "Rendering", "Default Scale"

       "``pos``","A real number between 0 and 1","Spatial position on the axis","**No default**"
       "``line_style``","String.","Style of line to use (equivalent to nicknames passed to the :meth:`~rayon.toolbox.ToolBox.new_line` method.","``solid``"
       "``line_width``,"Number","Width of line in points/pixels","``1``"
       "``line_color``","A color specification (see :ref:`specifying-colors`)","Color of line","``#000000FF``"
       "``ticksets``","List",":ref:`Tick sets <ref-ticks>` to draw on the line","Empty list"
    """
    
    axes = ('pos', 'line_style', 'line_width', 'line_color', 'ticksets')
    default_scales=dict(line_style='solid', line_width=1,
                        line_color="black", ticksets=[])
    @classmethod
    def from_border(cls, border, tickset_idx=0, **addl_params):
        """
        Create a :class:`GridLines` object from a  tickset in *border*, a
        :class:`~borders.TickableBorder`. By default, the first (or
        only) tickset associated with the border will be used.
        """
        tickset = border.get_tickset(tickset_idx)
        return cls(pos_data=[p for p in tickset.iter_tick_positions()],
                   pos_scale=tickset.get_scale(),
                   **addl_params)

    @classmethod
    def from_tick_tuples(cls, scale, tick_tuples, **addl_params):
        """
        Create a :class:`GridLines` object from a set of tick tuples
        and a scale.
        """
        return cls(pos_data=[x for x, y in tick_tuples],
                   pos_scale=scale,
                   **addl_params)
                   

    
    def __init__(self, **plot_kwargs):
        if type(self) == GridLines:
            raise RayonException("GridLines cannot be instantiated.")
        # Check keyword args for axis labels. If they exist, assume
        # they're trying to set that axis's scale to a constant, and
        # munge accordingly
        for k, v in plot_kwargs.items():
            if k in self.axes:
                del plot_kwargs[k]
                plot_kwargs["%s_scale" % k] = v
        plots.Plot.__init__(self, **plot_kwargs)

    def draw_(self, ctx, width, height):
        line_cache = {}
        def get_line(style, width, color):
            line_args = (style, width, color)
            if not line_cache.has_key(line_args):
                line_cache[line_args] = lines.create(style, color=color,
                                                     width=width)
            return line_cache[line_args]
        for pos, lstyle, lwidth, lcolor, ticksets in self.get_scaled_points():
            line = get_line(lstyle, lwidth, lcolor)
            self.do_draw(ctx, width, height, pos, line, ticksets)
    def do_draw(self, ctx, width, height, pos, line, ticksets):
        assert self.__class__.__name__ != "Gridlines", (
            "do_draw must be implemented in the %s" % (
            self.__class__.__name__))

class VerticalGridLines(GridLines):
    # documented in classref-gridlines
    def do_draw(self, ctx, width, height, x, line, ticksets):
        GridLines.do_draw(self, ctx, width, height, x, line, ticksets)
        scaled_x = x * width
        line.draw(ctx, scaled_x, 0, scaled_x, height)
        if ticksets is not None:
            for ticks in ticksets:
                ticks.draw_vertically(ctx, width, height, scaled_x)

class HorizontalGridLines(GridLines):
    # documented in classref-gridlines
    def do_draw(self, ctx, width, height, y, line, ticksets):
        GridLines.do_draw(self, ctx, width, height, y, line, ticksets)
        scaled_y = height - (y * height)
        #print "LINE: %s, %s, %s, %s" % (0    , scaled_y,
        #                                width, scaled_y)
        line.draw(ctx, 0, height - (y * height), width, scaled_y)
        if ticksets is not None:
            for tickset in ticksets:
                tickset.draw_horizontally(ctx, width, height, scaled_y)
        


_gridlines_classes = {
    'horizontal' : HorizontalGridLines,
    'vertical'   : VerticalGridLines
}

def create(gl_type, **kargs):
    return _gridlines_classes[gl_type](**kargs)

def create_from_border(gl_type, border, tickset_idx=0, **kargs):
    return _gridlines_classes[gl_type].from_border(border, tickset_idx,
                                                 **kargs)

def create_from_tick_tuples(gl_type, scale, tick_tuples, **kargs):
    return _gridlines_classes[gl_type].from_tickset(border, scale, tick_tuples,
                                                  **kargs)

def list_gridlines_types():
    return tuple((k, v) for k, v in _gridlines_classes.items())
