#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
"""
Tools for handling raster image data (i.e., colored-pixel bitmaps)
"""

from __future__ import division

import array
import sys
import itertools


def premultiply_rgba(pixel):
    try:
        a = pixel[3]
    except IndexError:
        a = 255
    r, g, b = tuple(int(x * a/255) for x in pixel[:3]) 
    return r, g, b, a

def unpremultiply_rgba(pixel):
    try:
        a = pixel[3]
    except IndexError:
        a = 255
    if a == 0:
        return pixel
    else:
        r, g, b = tuple(int(x / (a/255)) for x in pixel[:3])
        return r, g, b, a



class Formatter(object):
    """
    A class representing how pixels are formatted in a :class:`Bitmap`
    object. In general, users do not create instances of this
    class. Instead, they use one of the predefined :class:`RGBA32`,
    :class:`RGBA32_PM`, :class:`ARGB32`, :class:`ARGB32_PM`,
    :class:`BGRA32` or :class:`BGRA32_PM` instances.
    """
    def __init__(self, name, to_format, from_format):
        self._to_format = to_format
        self._from_format = from_format
        self._name = name
    def to_format(self, pixel):
        # Takes a pixel in RGBA32, returns a pixel in the specified
        # format
        return self._to_format(pixel)
        pass
    def from_format(self, pixel):
        # Takes a pixel in the specified format, returns a pixel in
        # RGBA32
        return self._from_format(pixel)
    def __repr__(self):
        return "<%s Formatter object at 0x%x>" % (self._name, id(self))


# Conversion functions to_ functions take unpremultiplied RGBA values
# and return whatever they return. from_ functions take whatever they
# take and return unpremultiplied RGBA.

# RGBA

RGBA32 = Formatter("RGBA32", lambda px: px, lambda px: px)
RGBA32_PM = Formatter("RGBA32_PM", premultiply_rgba, unpremultiply_rgba)


# ARGB

def to_argb32(px):
    r, g, b, a = px
    return a, r, g, b

def from_argb32(px):
    a, r, g, b = px
    return r, g, b, a

ARGB32 = Formatter("ARGB32", to_argb32, from_argb32)

def to_argb32_pm(px):
    return to_argb32(premultiply_rgba(px))

def from_argb32_pm(px):
    return unpremultiply_rgba(from_argb32(px))

ARGB32_PM = Formatter("ARGB32_PM", to_argb32_pm, from_argb32_pm)


# BGRA

def to_bgra32(px):
    r, g, b, a = px
    return b, g, r, a

def from_bgra32(px):
    b, g, r, a = px
    return r, g, b, a

BGRA32 = Formatter("BGRA32", to_bgra32, from_bgra32)

def to_bgra32_pm(px):
    return to_bgra32(premultiply_rgba(px))

def from_bgra32_pm(px):
    return unpremultiply_rgba(from_bgra32(px))
    

BGRA32_PM = Formatter("BGRA32_PM", to_bgra32_pm, from_bgra32_pm)





def transcode_pixel(pixel, in_fmt, out_fmt):
    # Turn a pixel in *in_fmt* into a pixel in *out_fmt)
    if in_fmt == out_fmt:
        return pixel
    else:
        return out_fmt.to_format(in_fmt.from_format(pixel))



class Bitmap(object):
    """
    An array of 4-byte words representing a 2-dimensional image. Each
    word represents a pixel; the :class:`Formatter` object associated
    with the object controls how this is done. Additional formatters
    determine the format of pixels as they are modified or read from
    the :class:`Bitmap` object.

    *width* and *height* are the width and height of the image in pixels.

    *fmt* is a :class:`Formatter` object specifying the default
    format of pixel input.

    By default, the initial value of the pixels in the bitmap is ``(0,
    0, 0, 0)``. This can be adjusted in one of two ways.

    If supplied, *initializer* is a function that a takes numeric
    index representing the pixel's position in the array, and returns
    a 4-tuple of integers between 0 and 255 to use as an initial value
    for that pixel.

    Alternately, *buff* is an :class:`array.array` object that the
    :class:`Bitmap` object will use as its internal image buffer. The
    :class:`Bitmap` object should be considered the buffer's "owner"
    and will modify the array as necessary. In this case, the user is
    responsibile for ensuring the array is an array of bytes of the
    appropriate length, e.g.::

       array.array("B", (0 for _ in xrange(width * height * 4)))

    where ``width`` and ``height`` are the *width* and *height*
    parameters passed to the constructor.

    If both *buff* and *initializer* are supplied, only *buffer*
    will be used.
    """
    def __init__(self,
                 width,
                 height,
                 fmt=None,
                 initializer=None,
                 buff=None):
        self._width = width
        self._height = height
        if fmt is None:
            self._fmt = RGBA32
        else:
            if not isinstance(fmt, Formatter):
                raise ValueError("fmt must be valid Formatter")
            self._fmt = fmt

        if buff is not None:
            self._buff = buff
        elif initializer is not None:
            self._buff = array.array(
                "B",
                itertools.chain.from_iterable(
                    (initializer(i) for i in xrange(width * height))))
        else:
            self._buff = array.array(
                "B", (0 for _ in xrange(width * height * 4)))

    def set_pixel(self, index, pixel, fmt=None):
        """
        Sets the pixel at *index* to the value specified by
        *pixel*. If *pixel* is in a format other than the default for
        this object, *fmt* specifies it.
        """
        if fmt is not None:
            # Convert pixel to internal format
            pixel = transcode_pixel(pixel, fmt, self._fmt)
        index = index * 4
        for offset, val in enumerate(pixel):
            self._buff[index+offset] = val
            
    def get_pixel(self, index, fmt=None):
        """
        Returns the pixel at *index*. If *fmt* is specified, return it
        in the requested format. Otherwise, return it in the default
        format of the bitmap.
        """
        pixel = tuple(self._buff[index * 4 : (index + 1) * 4])
        if fmt is not None:
            return transcode_pixel(pixel, self._fmt, fmt)
        else:
            return pixel

    def get_width(self):
        "Returns the width of the bitmap in pixels."
        return self._width
    
    def get_height(self):
        "Returns the height of the bitmap in pixels."
        return self._height

    def get_format(self):
        """
        Returns the default :class:`Formatter` object associated with
        this object.
        """
        return self._fmt

    def get_data(self, fmt=None):
        """
        Returns a buffer containing the pixel data in this map. Pixel
        data will be in the object's default pixel format unless *fmt*
        is specified.

        If *fmt* is unspecified or equal to the default format for the
        object, the buffer returned will be the same buffer used to
        hold the pixel data internally. As with so many of history's
        greatest mistakes, this is done for efficiency.
        """
        if fmt is None or fmt == self._fmt:
            return self._buff
        else:
            gen = (j
                   for i in xrange(0, len(self._buff), 4)
                   for j in transcode_pixel(self._buff[i:i+4],
                                            self._fmt,
                                            fmt))
            return array.array("B", gen)


    def upscale(self, factor):
        """
        Returns a bitmap identical to this one, except that each pixel
        has been copied *factor* times in both width and height. Each
        pixel in the source bitmap is mapped into a square of *factor*
        pixels per side in the in the result bitmap.

        This method should not be used to do general-purpose upscaling
        of bitmaps intended for display. It can be useful when a
        :class:`Bitmap` object is being used as a container for data,
        and for certain types of images (e.g., small images containing
        only vertical and horizontal edges), which the renderer may
        not scale well.
        """
        def mult_iter():
            for row_offset in xrange(0, len(self._buff), self._width * 4):
                # Do each row *factor* times
                for _ in xrange(factor):
                    for px_offset in xrange(row_offset,
                                            row_offset + (self._width * 4),
                                            4):
                        # Do each pixel *factor* times
                        for _ in xrange(factor):
                            yield self._buff[px_offset:px_offset+4]
        new_buff = array.array("B", itertools.chain.from_iterable(mult_iter()))
        return Bitmap(width=self._width * factor,
                      height=self._height * factor,
                      fmt=self._fmt,
                      buff=new_buff)
                      
                      

                
        

    def is_old(self):
        # (Undocumented) Used to tell the difference between this
        # object and the old Bitmap object in backgrounds.
        return False

    # Sequence protocol stuff
    def __getitem__(self, idx):
        """
        Equivalent to calling the :meth:`get_pixel` method without
        supplying a :class:`Formatter` object. (i.e., the returned
        pixel will be in the default format.)
        """
        return self.get_pixel(idx)

    def __setitem__(self, idx, value):
        """
        Equivalent to calling the :meth:`set_pixel` method without
        supplying a :class:`Formatter` object. (i.e., the *value* is
        presumed to be a pixel in the :class:`Bitmap` object's default
        format.)
        """
        return self.set_pixel(idx, value)

    def __len__(self):
        """
        Returns the number of pixel elements stored in this
        bitmap. (i.e., the size of the underlying :class:`array.array`
        object divided by 4)
        """
        return len(self._buff) / 4


