#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
from __future__ import division
"""
Utility functions to make writing Rayon command-line tools easier.

This library defines code for specifying what options exist for a
command-line tool. The :func:`execute` function uses a command's
options and configuration to create a sane environment, then calls the
command's entry point.

The following is a minimal but complete example of use::

   from rytools import *
   
   def main(options):
       if options['guess'] == 1:
           print "You guessed it, %s!" % options['user']
       elif options['guess'] < 1:
           print "Too low, %s" % options['user']
       else:
           print "Too high, %s" % options['user']

   opts = [StringOption("user"),
           IntOption("guess")]

   execute(cmd_func=main, cmd_name="guesser", cmd_options=opts)

Command Name
------------

The name of the command is passed to the :func:`execute` function via
the ``cmd_name`` argument. If it is not supplied, the value of
``sys.argv[0]`` will be used as the command name.

Options
-------

A command specifies a collection of options, which may be supplied
either at the command line or in a configuration
file. :class:`StringOption` objects return the value as presented in
the configuration, as a string; other option types convert the value
in some way to a value useful to the command. Implementors may
subclass the :class:`Option` object to apply their own semantics to
configuration information.


Configuration
-------------

Options may be specified on the command line or from configuration
files. Configuration is applied in the following order:

   * System-wide configuration is read from ``/etc/ryrc``, if it
     exists.

   * Per-user configuration is read from ``~/.ryrc``, if it exists.

   * Per-environment configuration is read from the file pointed to by
     the ``RYRC`` environment variable, if it is set.

   * Per-command configuration is read from the configuration file
     provided by the ``--config-file`` command-line parameter, if it
     is supplied.

   * Per-command configuration is read from the command-line
     parameters.

The configuration file format is as described in the
:mod:`ConfigParser` module. When each file is read, configuration
information is taken first from the ``ALL`` section of the file, and
then from a section whose name corresponds to the command being
executed (i.e., the value of the ``cmd_name`` to the :func:`execute`
function). ``ALL`` can therefore be used for configuration information
that applies to all Rayon tools.

If a configuration file (or the command line) defines an
``ignore_config`` directive (``--ignore-config`` on the command line)
in one of the parsed sections, all configuration information gathered
prior to the parsing of that config file or command line will be
discarded. This may be used to ignore variable in system or user
config files for specific tool invocations, for instance.

Conversion Functions
--------------------

The :mod:`rytools` module provides several common conversion
functions. These may be used as the value of the *converter* parameter
of the :class:`Option`, :class:`ListOption` and :class:`KVOption`
classes.

   .. function:: colname_from_str(s: str)

      Parses a string representation of a column name in a Rayon
      :class:`~rayon.data.Dataset` object. The input may be either
      column names or numeric indices. (If the dataset has been passed
      into the tool as a file, as from the ``--input-path`` option,
      columns are numbered left to right starting at 0.)

   .. function:: color_from_str(s: str)
   
      Converts a string from the command line to a color tuple using
      the rules in :ref:`specifying-colors`. Because it parses
      command-line input and shells often use the octothorpe (``#``)
      as a comment character, it may be omitted in the input to this
      function.

   .. function:: ip_from_str(s: str)

      Converts an IPv4 host address to an integer. The address can be
      expressed either as an integer string (in which case conversion
      is straightforward) or in "dotted-quad" notation. 

   .. function:: istream_from_param(
"""
import os
import sys
import errno
import optparse
import ConfigParser
import struct
import socket


from pprint import pprint

import rayon # for version
from rayon.cliutils import mk_tickspec_parser
from rayon.tickspec_parse import parse_tickspec
import rayon.miscutils as miscutils

from netsa.data.times import make_datetime, make_timedelta

def normpath(p):
    if p is None:
        return None
    return os.path.normpath(os.path.expanduser(p))

class ToolRuntimeError(Exception):
    """
    Indicates that the tool identified unacceptable runtime
    behavior. The error message is expected to be output to a user at
    a command-line console.
    """
    pass

class ToolConfigurationError(Exception):
    """
    Indicates that something is wrong with the configuration options
    for the command. The error message is expected to be output to a
    user at a command-line console.
    """
    pass

class Constant(object):
    def __init__(self, name): self._name = name
    def __repr__(self): return "Constant('%s')" % self._name

NO_DEFAULT = Constant("NO_DEFAULT")

SUPPRESS_HELP = Constant("SUPPRESS_HELP")


### Option Reboot

def _extract_name(n):
    if n.startswith("--"): return n[2:]
    else:                  return n
def _extract_valname(n, v):
    if v is None: return n.replace("-", "_")
    else:         return v

class ConfigOption(object):
    pass


# TODO: Should default value be taken but ignored??

class _Flag(ConfigOption):
    metavar = "FLAG"
    truth = True
    def __init__(self, optname, valname=None, hlp=SUPPRESS_HELP):
        optname = _extract_name(optname)
        valname = _extract_valname(optname, valname)
        # The "default value" is the value of this flag when it's not
        # supplied, hence: !self.truth
        self._vals = (optname, valname, not self.truth, hlp, self.metavar)
    def get_option_name(self): return self[0]
    def get_value_name(self): return self[1]
    def get_default_value(self): return self[2]
    def get_help(self): return self[3]
    def get_metavariable(self): return self[4]
    def __getitem__(self, key):
        return self._vals[key]
    def __call__(self, val):
        # Note that call should only be used when the Flag is to be
        # treated like any other option. (I.e., when it's not being
        # used as a flag, but as a config value with a boolean value
        # type, as in a config file.) When it's being used as a flag
        # (i.e., at the command line), it should be treated as a flag
        # in the code (that is, check if it exists and if it does,
        # treat it as True).
        if isinstance(val, str) and (val.lower() in ("0", "false")):
            return not self.truth
        elif val in (0, False):
            return not self.truth
        else:
            return self.truth
    def __repr__(self):
        return "%s%s" % (type(self).__name__, str(self._vals))

    
class Flag(_Flag):
    """
    A positive unary option. A flag is written on the command line as
    an option with no arguments, (``--foo``). In a configuration file,
    the values ``0`` and the case-insensitive string ``false`` set a
    flag to ``False``; all other values set the flag to ``True``.
    """
    pass


class FalseFlag(_Flag):
    """
    A negative unary option. When you want to express the absence of a
    configuration option (e.g., ``--dont-render-png``) in a positive way
    (``render_png_enabled``), use a :class:`FalseFlag`.

    A false flag has the opposite semantics to a :class:`Flag`
    object. Specifying the option on the command line or configuration
    file sets the underlying option value to ``False``; omitting it
    sets it to ``True``. (Note that setting a :class:`FalseFlag` to
    ``false`` or ``0`` in a configuration file will set the underlying
    option value to ``True``.)
    """
    truth=False

    
class Option(ConfigOption):
    """
    Base class for options.

    *optname* is the name of the option as it is used in the
    configuration file or on the command line.

    If specified, *valname* is the key in which to store the value of
    this option in the parsed configuration. By default this is the
    same as *optname*.

    If specified, *default_val* is the default value given for this
    option when it is not supplied. By default an option is required,
    and an error will be raised if it is not supplied. This is the
    default behavior.

    If specified, *hlp* is help text to display as part of the tool
    usage. If this is not supplied an option will not be documented as
    part of the usage output.

    If specified, *converter* is a function that takes a string and
    returns a value of the expected type for this option. By default,
    option values are stored as strings.
    """
    metavar_ = "ARG"
    def __init__(self,
                 optname,
                 valname=None,
                 default_val=NO_DEFAULT,
                 hlp=SUPPRESS_HELP,
                 converter=lambda s: s):
        optname = _extract_name(optname)
        valname = _extract_valname(optname, valname)
        self._vals = (optname, valname, default_val, hlp, self.metavar_)
        self._converter = converter
    def get_option_name(self): return self[0]
    def get_value_name(self): return self[1]
    def get_default_value(self): return self[2]
    def get_help(self): return self[3]
    def get_metavariable(self): return self[4]
    def __getitem__(self, key):
        return self._vals[key]
    def __call__(self, val):
        return self._converter(val)
    def __repr__(self):
        return "%s%s" % (type(self).__name__, str(self._vals))


def _split(s, delim):
    # If you get an empty string, return an empty
    # list. Dammit.
    if len(s) == 0: return tuple()
    else:           return tuple(i.strip() for i in s.split(delim))


class ListOption(Option):
    """
    A :class:`Option` object that takes a delimited list of values and
    converts them into a Python tuple.

    A :class:`ListOption` object takes the following parameter in
    addition to those of the :class:`Option` object:

    If supplied, *delimiter* the character that will be used to
    separate items in the list. The default delimiter is ``,``.
    """
    metavar_ = "ARGS"
    def __init__(self,
                 optname,
                 varname=None,
                 default_val=NO_DEFAULT,
                 hlp=SUPPRESS_HELP,
                 converter=lambda x: x,
                 delimiter=","):
        if delimiter in (None, ""):
            raise ValueError("Delimiter must be a non-empty string")
        Option.__init__(
            self,
            optname,
            varname,
            default_val,
            hlp,
            lambda val: tuple(converter(i.strip())
                              for i in _split(val, delimiter)))
        
        self._vals = self._vals + (delimiter,)


class ChoiceOption(Option):
    """
    A :class:`Option` object that takes one of a preconceived set of
    values.

    A :class:`ChoiceOption` object takes the following parameter in
    addition to those of the :class:`Option` object:

    *choices* is a list of possible values for this option. The
    choices should be in post-converted form. (For example, if
    *converter* is ``int``, ``(1, 2, 3)`` is a valid value for
    *choices*; ``("1", "2", "3")`` is not.)
    """
    metavar_ = "ARGS"
    def __init__(self,
                 optname,
                 varname=None,
                 default_val=NO_DEFAULT,
                 hlp=SUPPRESS_HELP,
                 converter=lambda x: x,
                 choices=None):
        if choices is None or len(choices) == 0:
            raise ToolConfigurationError(
                "%s: Can't create ChoiceOption with empty choice list"
                % optname)

        choices = tuple(choices)
        
        def check_and_convert(val):
            val = converter(val)
            if val not in choices:
                raise ToolRuntimeError(
                    "'%s' not a valid value for '%s'" % (
                        val, optname))
            return val

        Option.__init__(
            self,
            optname,
            varname,
            default_val,
            hlp,
            check_and_convert)

        self._vals = self._vals + (choices,)
        

class KVOption(Option):
    """
    A :class:`Option` object that takes a delimited list of key/value
    pairs, and converts them into a tuple of pairs. (This can be
    passed into a dictionary constructor to make a dictionary if so
    desired.)

    A :class:`KVOption` object takes the following parameter in
    addition to those of the :class:`Option` object:

    If supplied, *converter* is applied to the value of each key-value
    pair. By default, values are returned as strings.

    If supplied, *delimiter* the character that will be used to
    separate items in the list. The default delimiter is ``,``.

    If supplied, *kv_delimiter* is the character that separates keys
    from values within each item. The default key-value delimiter is
    ``=``.
    """
    metavar = "ARGS"
    def __init__(self,
                 optname,
                 varname=None,
                 default_val=NO_DEFAULT,
                 hlp=SUPPRESS_HELP,
                 converter=lambda x: x,
                 delimiter=",",
                 kv_delimiter="="):
        if delimiter in (None, ""):
            raise ValueError("delimiter must be a non-empty string")
        if kv_delimiter in (None, ""):
            raise ValueError("pair delimiter must be a non-empty string")
        Option.__init__(
            self,
            optname,
            varname,
            default_val,
            hlp,
            lambda val: tuple((k.strip(), converter(v.strip()))
                              for k, v in (i.split(kv_delimiter)
                                           for i in _split(val, delimiter))))
        
        self._vals = self._vals + (delimiter, kv_delimiter)


# Several convenience classes to do common single-value conversions.

# -- String and Path options don't actually convert their input.

class StringOption(Option):
    """
    An option whose value is a string.
    """
    metavar_="STRING"

class PathOption(Option):
    """
    An option whose value is a path name.
    """
    metavar_="PATH"


# -- Everything else operates on its input in some
# -- way. _ConverterOption takes care of properly calling the
# -- superclass to make this happen.

class ConverterOption(Option):
    def __init__(self,
                 optname,
                 valname=None,
                 default_val=NO_DEFAULT,
                 hlp=SUPPRESS_HELP):
        return Option.__init__(self,
                                  optname,
                                  valname=valname,
                                  default_val=default_val,
                                  hlp=hlp,
                                  converter=self.converter_)

class IntOption(ConverterOption):
    """
    An option whose value is an integer.
    """
    metavar_="INT"
    converter_=int

class FloatOption(ConverterOption):
    """
    An option whose value is a floating-point number.
    """
    metavar_="FLOAT"
    converter_=float

# Tick specifications:

# Deprecated: Use NewTickspecOption
class TickspecOption(ConverterOption):
    metavar_="TICKSPEC"
    converter_ = staticmethod(mk_tickspec_parser)

class NewTickspecOption(ConverterOption):
    """
    An option whose value is a tick specification.
    """
    metavar_="TICKSPEC"
    def converter_(self, v):
        return parse_tickspec(self.get_option_name(), v)
        

# Column names:

def colname_from_str(s):
    try:               return int(s)
    except ValueError: return str(s)
    

class ColnameOption(ConverterOption):
    """
    An option whose value is a column name or index in a
    :class:`~rayon.data.Dataset` object.
    """
    metavar_="COLNAME_OR_IDX"
    converter_=staticmethod(colname_from_str)


# Colors:

def color_from_str(s):
    try:
        return miscutils.parse_color_string(s)
    except ValueError, e:
        if s.startswith("#"):
            raise e
        else:
            try:
                return miscutils.parse_color_string("#" + s)
            except:
                raise e

class ColorOption(ConverterOption):
    """
    An option whose value is a color specification. (See
    :ref:`specifying-colors`). The value will be converted to either a
    3- or 4-tuple RGB(A) value. (If not specified, the value for alpha
    is 255.)
    """
    metavar_="COLOR"
    converter_=staticmethod(color_from_str)

            

# IP Addresses:

def ip_from_str(s):
    try:
        return int(s)
    except ValueError:
        try:
            return struct.unpack("!L", socket.inet_aton(s))[0]
        except socket.error, e:
            raise ValueError("%s: %s" % (s, str(e)))

class IpOption(ConverterOption):
    """
    An option whose value is a "dotted-quad" or integer representation
    of an IPv4 address. The value will be converted to an integer.
    """
    metavar_="INT_OR_IP"
    converter_=staticmethod(ip_from_str)

# Dates and times

class DateOption(ConverterOption):
    """
    An option whose value is a date. Valid input is identical to that
    of :func:`netsa.data.times.make_datetime`. The value will be
    converted to a :class:`datetime.datetime` object.
    """
    metavar_="DATE"
    converter_=staticmethod(make_datetime)

class TimeDeltaOption(ConverterOption):
    """
    An option whose value is an ISO-8601-compatible time delta. The
    value will be converted to a :class:`datetime.timedelta` object.
    """
    metavar_="TIMEDELTA"
    converter_=staticmethod(make_timedelta)

    
class _NotSpecified(Exception):
    pass

def _merge_with_defaults(opts, parsed):
    # Merge the parsed output in *parsed* with the default values in
    # opts
    defaults = ((o.get_value_name(), o.get_default_value())
                for o in opts
                if o.get_default_value() != NO_DEFAULT)
    required = set(o.get_value_name()
                   for o in opts
                   if o.get_default_value() == NO_DEFAULT)
    merged = dict(defaults)
    merged.update(parsed)
    # Check that all the required arguments are here. If not, that's
    # an error.
    for r in required:
        if r not in merged:
            raise _NotSpecified(r)
    return merged



#
# Functions to parse configuration/command-line into dictionaries
#


def parse_config_file(options, cfg, section, no_sections=False):
    """
    Extract configuration information from config files.

    *options* is an iterable of :class:`Option` objects.

    *cfg* is a :class:`ConfigParser.ConfigParser` object

    By default, this function will get configuration data from the
    ``ALL`` section of the config file, then a call-specific
    section named by *section*.

    If *no_sections* is ``True``, get all values from the ``DEFAULT``
    section instead. (And ignore ``ALL`` and *section*.)

    This function does _not_ do the following:

       * No type conversion is performed in this function.
       * Default values are not applied
       * Missing data is not detected.

    Unexpected options are detected and will cause an error, unless
    the unexpected option is in the ``ALL`` section. In that case, it
    will be ignored.

    This function returns a dictionary containing the configuration
    information found in the configuration file.
    """
    options_byname = dict((o.get_option_name(), o) for o in options)
    def dict_from_cfg_section(s):
        if s != "DEFAULT" and not cfg.has_section(s):
            return dict()
        rc = dict()
        for k, v in cfg.items(s):
            if k not in options_byname:
                if s == 'ALL':
                    # ignore unexpected keys in ALL
                    continue
                raise ToolConfigurationError(
                    "Unexpected argument in section %s: %s" % (s, k))
            o = options_byname[k]
            try:
                rc[o.get_value_name()] = v
            except:
                raise ToolConfigurationError(
                    "Error converting '%s' to a '%s'" % (
                        v, o.get_metavariable()))

        return rc
        
    if no_sections:
        return dict_from_cfg_section('DEFAULT')
    else:
        rc = dict_from_cfg_section('ALL')
        rc.update(dict_from_cfg_section(section))
        return rc
        
                      


def parse_command_line(options,
                       args=None,
                       desc="",
                       debug=False):
    """
    Extract configuration information from the command line.

    *options* is an iterable of :class:`Option` objects.

    If specified, *args* is a list of words representing a set of
    arguments. By default, ``sys.argv[1:]`` is used.

    If specified, *desc* is a short description of the tool that will
    be printed as part of the usage message if command-line parsing
    fails.

    This function does _not_ do the following:

       * No type conversion is performed in this function.
       * Default values are not applied
       * Missing data is not detected.

    However, unexpected options are detected and will cause an error.

    This function returns a dictionary containing the configuration
    information found on the command line, and a tuple of positional
    arguments that were left over after processing.
    """
    # Subclass OptionParser to customize error-handling
    class _MyOptionParser(optparse.OptionParser):
        def error(self, msg):
            raise ToolConfigurationError(msg)
     
    #optparser = optparse.OptionParser(usage="%%prog [options] - %s" % desc)
    optparser = _MyOptionParser(usage="%%prog [options] - %s" % desc)
       

    options_byvalname = dict()

    for o in options:
        name, valname, default_val, hlp, metavar = o[:5]
        if hlp == SUPPRESS_HELP:
            hlp = optparse.SUPPRESS_HELP
        if isinstance(o, (Flag, FalseFlag)):
            # In this case, True just means the option was seen, so it
            # doesn't matter if the option type was Flag or FalseFlag;
            # this value will be merged with config files and have its
            # truth value adjusted appropriately.
            optparser.add_option("--%s" % name,
                                 dest=valname,
                                 metavar=metavar,
                                 help=hlp,
                                 action="store_true")
        else:
            optparser.add_option("--%s" % name,
                                 dest=valname,
                                 metavar=metavar,
                                 help=hlp,
                                 type="str")
        options_byvalname[valname] = o
        

    # (Don't let optparse create a Values object for us, because it
    # will contain an entry for every option, even if it isn't
    # supplied. We only want the ones present on the command line.)
    parsed_options, positional_args = optparser.parse_args(
        args, values=optparse.Values())

    return vars(parsed_options), tuple(positional_args)


def merge_configuration(options, *configs):
    """
    Returns a dictionary containing merged configuration information
    from *configs*, supplying default values from *options*. If a
    value is missing and does not have a default value in *options*,
    an error is raised.

    If any dictionary defines the key ``ignore_config``, configuration
    information from previously parsed dictionaries is thrown out.

    *options* is an iterable of :class:`Option` objects.
    
    *configs* is a sequence of dictionaries. The dictionaries will be
    merged in the order they are provided; if keys are repeated,
    values in later dictionaries will overwrite values in earlier
    dictionaries, e.g.::

       >>> o = [IntOption('foo'),
       ...      IntOption('bar'),
       ...      IntOption('baz')]
       ...
       >>> a = {'foo': 1, 'bar': 2}
       >>> b = {'foo': 3, 'baz': 4}
       >>> merge_configuration(o, a, b)
       >>> {'baz': 4, 'foo': 3, 'bar': 2}

    *configs* is presumed not to contain unexpected values (i.e., keys
    that aren't value names in *options*). The
    :func:`parse_config_file` and :func:`parse_command_line` functions
    guarantee this.
    """
    rc = {}
    for c in configs:
        if c.has_key('ignore_config'):
            rc = dict(c)
            del rc['ignore_config']
        else:
            rc.update(c)

    # NOTE: Unexpected values should be handled by parse_config_file
    # and parse_command_line, which provide our input dictionaries, so
    # no checking for unexpected values is done here.

    # Provide default values for missing options. Convert non-missing
    # options to their target types.
    for o in options:
        (name, valname, default) = o[:3]
        if valname not in rc:
            # If default is NO_DEFAULT, this gets baked into rc. The
            # execute() method checks for options with a value of
            # NO_DEFAULT and raises an error if they have gotten
            # through.
            rc[valname] = default
        else:
            rc[valname] = o(rc[valname])

    return rc




#
# The execute function and its helpers
#

    
        
# Default options

default_width  = 400
default_height = 400

# TODO: the sytem_options/common_options distinction may be
# superfluous. Eliminate?
#
# system_options == options controlling configuration parsing
#
# common_options == options unrelated to config parsing that all apps
# "should have"

system_options = [
    Flag("ignore-config", 
         hlp="Ignore any configuration information from configuration files"),
    StringOption("config-file",
                 default_val=None,
                 hlp="File to use for visualization configuration"),
    Flag("print-options"),
    Flag("version")
]

common_options = [
    PathOption(
        "input-path", default_val="-",
        hlp="Path to input data. (Default: STDIN)"),
    PathOption(
        "output-path", default_val=NO_DEFAULT,
        hlp="(REQUIRED) Path to output file"),
    IntOption(
        "width", default_val=default_width,
        hlp="image width (pixels for PNG, otherwise points) "
        "(Default: %d)" % default_width),
    IntOption(
        "height", default_val=default_height,
        hlp="image height (pixels for PNG, otherwise points) "
        "(Default: %d)" % default_height),
    Flag(
        "first-line-colnames", 
        hlp="First line of input contains column names")
]

        



def coalesce_configuration(command,
                           args_list,
                           cmd_options,
                           cmd_desc,
                           files_to_read,
                           envars):
    """
    Figure out which configuration sources to read, read them, and
    merge them into a single configuration dictionary.

    *command* is the name of the command being run.

    *args_list* is the command-line argument list, excluding the
    command name (e.g., ``sys.argv[0]``)

    *cmd_options* is a sequence of :class:`Newoption` objects.

    *cmd_desc* is a string describing the command

    *files_to_read* is a sequence of file names representing
    configuration files in the order in which they should be read.

    *envars* is a set of environment variable names that may contain
    configuration file names. These will be read and applied after the
    files in *files_to_read*, in the order specified in *envars*.
    """
    if cmd_options   is None: cmd_options   = tuple()
    if cmd_desc      is None: cmd_desc      = ""
    if files_to_read is None: files_to_read = tuple()
    if envars        is None: envars        = tuple()


    def filename_from_envar(envar):
        if (os.getenv(envar) not in (None, "")
            and os.path.exists(os.getenv(envar))):
                return os.getenv(envar)
        else:
            return None

    envar_files_to_read = tuple(f for f in (filename_from_envar(e)
                                            for e in envars)
                                if f is not None)

    # Parse command line now to get --config-file
    cli_dict, leftovers = parse_command_line(cmd_options,
                                             args=args_list,
                                             desc=cmd_desc)
    
    if "config_file" in cli_dict:
        cfg = ConfigParser.ConfigParser()
        if cli_dict['config_file'] == "-":
            # Special handling
            cfg.readfp(sys.stdin, 'STDIN')
        else:
            cfgpath = (
                os.path.normpath(
                    os.path.expanduser(
                        cli_dict['config_file'])))
            if not os.path.exists(cfgpath):
                raise ToolConfigurationError("%s: no such file or directory" %
                                             cfgpath)
            cfg.read([cfgpath])
        cli_config_dict = parse_config_file(cmd_options, cfg, command)
    else:
        cli_config_dict = {}
        
    def config_dict_from_file(fname):
        if os.path.exists(fname):
            cfg = ConfigParser.ConfigParser()
            cfg.read([fname])
            rc = parse_config_file(cmd_options, cfg, command)
            return rc
        else:
            return {}

    config_dicts = tuple(config_dict_from_file(f)
                         for f in (files_to_read + envar_files_to_read))

    # print "-" * 70
    # pprint(config_dicts)
    # print "-" * 70
    

    all_dicts = config_dicts + (cli_config_dict, cli_dict)
    merged = merge_configuration(cmd_options, *all_dicts)
    # Add unparsed arguments from command line as __EXTRA__
    merged['__EXTRA__'] = leftovers
    return merged


def execute(cmd_name=None, cmd_func=None, cmd_args=None,
            cmd_options=None, include_common_options=True, cmd_desc=None,
            debug=False):
    """
    Runs the main loop in an rytools application.

    *cmd_name* is the name of the tool. If nothing is provided,
    ``sys.argv[0]`` will be used.

    *cmd_func* is a function which takes a single argument -- a
    dictionary of command options. It optionally returns an integer
    that will be used as an exit code. (If nothing is returned,
    success is assumed.)

    *cmd_args* is a list to use as the argument list. If not supplied,
    ``sys.argv[1:] will be used.

    *cmd_options* is a list of :class:`Option` objects that define the
     options the application wwill accept. If *include_common_options*
     is ``True`` (the default), the following base options will be
     added to this list:

        * ``input-path``
        * ``output-path``
        * ``width``
        * ``height``
        * ``first-line-colnames``

    The following options will be added regardless of the status of
    *include_common_options*:

       * ``ignore-config``
       * ``config-file``
       * ``help``

    To see the descriptions of these options, run any of the rytools
    with the ``--help`` switch.

    *cmd_desc* is a string description of the name. If not supplied,
    an empty string will be used.
    
    """
    if cmd_func is None:
        print "Not told to do anything. Exiting."
        sys.exit(0)

    if cmd_name is None:
        cmd_name = os.path.basename(sys.argv[0])

    if cmd_args is None:
        cmd_args = sys.argv[1:]

    if cmd_options is None:
        cmd_options = []

    if include_common_options:
        cmd_options = list(set(system_options) |
                           set(common_options) |
                           set(cmd_options))
    else:
        cmd_options = list(set(system_options) | set(cmd_options))

    # Check for duplicate options
    names, valnames = set(), dict()
    for o in cmd_options:
        name, valname = o[:2]
        if name in names:
            print "Error: option '%s' is defined twice" % name
            sys.exit(1)
        if valname in valnames:
            print ("Error: value name '%s' used in both options '%s' and '%s'\n"
                   % (valname, valnames[valname], name))
            sys.exit(1)
        names.add(name)
        valnames[valname] = name

    def maybe_enter_debugger(exc_info):
        if (os.getenv("RY_TRACE") is not None
            or os.getenv("RY_DBG") is not None                
            or os.getenv("RY_DEBUG") is not None
            or debug):
            import traceback
            traceback.print_exception(*exc_info)
        if os.getenv("RY_PDB") is not None:
            print str(e)
            import pdb
            pdb.post_mortem(exc_info[2])

    # Figure out what we're told
    try:
        opts = coalesce_configuration(
            command=cmd_name,
            args_list=cmd_args,
            cmd_options=cmd_options,
            cmd_desc=cmd_desc,
            files_to_read=('/etc/ryrc',
                           normpath("~/.ryrc")),
            envars=("RYRC",))

        # Special-case some options that can run regardless of
        # anything else.
        if opts.get("print_options", False):
            pprint(opts)
            return

        if opts.get("version", False):
            str_version = ".".join("%d" % i for i in rayon.version)
            print "%s: version %s" % (os.path.basename(sys.argv[0]), 
                                      str_version)
            return
        

        # Check that options with no defaults are provided
        for o in cmd_options:
            (name, valname, default) = o[:3]
            if opts[valname] == NO_DEFAULT:
                raise ToolRuntimeError(
                    "Required parameter '%s' not provided" % name)
    except Exception, e:
        print "Configuration error: %s" % str(e)
        print "Type '%s --help' for more information on arguments" % cmd_name
        maybe_enter_debugger(sys.exc_info())
        sys.exit(1)
    except KeyboardInterrupt:
        print "Application interrupted by user"
        sys.exit(1)


    # Do what we're told
    try:
        rc = cmd_func(opts)
        if rc is None:
            sys.exit(0)
        else:
            sys.exit(rc)
    except ToolRuntimeError, e:
        print "Error: %s" % str(e)
        maybe_enter_debugger(sys.exc_info())
        sys.exit(1)
    except Exception, e:
        print "Error (%s): %s" % (e.__class__.__name__, str(e))
        maybe_enter_debugger(sys.exc_info())
        sys.exit(1)
    except KeyboardInterrupt:
        print "Application interrupted by user"
        sys.exit(1)
        

# Some general utility functions

def istream_from_str(param):
    if param == "-":
        return sys.stdin
    else:
        return open(param, 'r')

istream_from_param = istream_from_str



__all__ = """
ChoiceOption
ColnameOption
ColorOption
Constant
DateOption
FalseFlag
Flag
FloatOption
IntOption
IpOption
KVOption
ListOption
NO_DEFAULT
NewTickspecOption
Option
PathOption
SUPPRESS_HELP
StringOption
TickspecOption
TimeDeltaOption
ToolConfigurationError
ToolRuntimeError
colname_from_str
color_from_str
execute
ip_from_str
istream_from_param
istream_from_str
""".split()

