#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@

import unittest

# Set this to True to test the actual Python 2.7+ OrderedDict
# instead. If you ever change these tests, make sure to test against
# the original OrderedDict to make sure you're testing for behavior
# that actually exists. :)
testing_real_ordered_dict = True

try:
    from collections import OrderedDict as coll_OrderedDict
except ImportError:
    # We're probably not running on a Python that supports
    # collections.OrderedDict. That is the purpose of the module,
    # after all.
    testing_real_ordered_dict = False

if testing_real_ordered_dict:
    OrderedDict = coll_OrderedDict
else:
    from rayon.backcompat import MyOrderedDict as OrderedDict

default_items = (('a', 1), ('b', 2), ('c', 3))

class TestReadOps(unittest.TestCase):
    d = OrderedDict(default_items)
    def test_brackets_accessor(self):
        self.assertEqual(1, self.d['a'])
    def test_len(self):
        self.assertEqual(3, len(self.d))
    def test_contains(self):
        self.assertTrue('a' in self.d)
        self.assertFalse('foo' in self.d)
    def test_has_key(self):
        self.assertTrue(self.d.has_key('a'))
        self.assertFalse(self.d.has_key('foo'))
    def test_keys(self):
        self.assertEqual([x[0] for x in default_items], self.d.keys())
    def test_iterkeys(self):
        self.assertEqual([x[0] for x in default_items], 
                         list(self.d.iterkeys()))
    def test_values(self):
        self.assertEqual([x[1] for x in default_items], self.d.values())
    def test_itervalues(self):
        self.assertEqual([x[1] for x in default_items], 
                         list(self.d.itervalues()))
    def test_items(self):
        self.assertEqual(list(default_items), self.d.items())
    def test_iteritems(self):
        self.assertEqual(list(default_items), list(self.d.iteritems()))
    def test_reversed(self):
        self.assertEqual(list(reversed([x[0] for x in default_items])), 
                         list(reversed(self.d)))
    def test_get_item_exists_nodefault(self):
        self.assertEqual(1, self.d.get('a'))
    def test_get_item_exists_default(self):
        self.assertEqual(1, self.d.get('a', 4))
    def test_get_item_not_exists_default(self):
        self.assertEqual(4, self.d.get('foo', 4))
    def test_get_item_not_exists_nodefault(self):
        self.assertEqual(None, self.d.get('foo'))
    def test_copy(self):
        d2 = self.d.copy()
        self.assertFalse(self.d is d2)
        self.assertEqual(self.d.items(), d2.items())
    def test_str(self):
        self.assertEqual(
            "OrderedDict([('a', 1), ('b', 2), ('c', 3)])",
            str(self.d))
    def test_repr(self):
        self.assertEqual(
            "OrderedDict([('a', 1), ('b', 2), ('c', 3)])",
            repr(self.d))

    def test_compare_succeeds(self):
        a = OrderedDict(default_items)
        self.assertTrue(a == self.d)
    def test_compare_fails_out_of_order(self):
        a = OrderedDict((('a', 1), ('c', 3), ('b', 2)))
        self.assertFalse(a == self.d)
    def test_compare_fails_missing_values(self):
        a = OrderedDict((('a', 1), ('b', 2)))
        self.assertFalse(a == self.d)
    def test_compare_fails_wrong_values(self):
        a = OrderedDict((('a', 1), ('b', 2), ('c', 4)))
        self.assertFalse(a == self.d)
        
    
    

class TestWriteOps(unittest.TestCase):
    # Initializer tests
    def test_init_no_args(self):
        d = OrderedDict()
        self.assertEqual(0, len(d))
    def test_init_from_empty(self):
        d = OrderedDict([])
        self.assertEqual(0, len(d))
    def test_init_from_tuples(self):
        d = OrderedDict(default_items)
        self.assertEqual(1, d['a'])
    def test_init_from_iter(self):
        d = OrderedDict(iter(default_items))
        self.assertEqual(1, d['a'])
    def test_init_from_dict(self):
        d = OrderedDict(dict(default_items))
        self.assertEqual(1, d['a'])
    def test_init_from_OrderedDict(self):
        d = OrderedDict(OrderedDict(default_items))
        self.assertEqual(1, d['a'])

    if testing_real_ordered_dict:
        def test_init_from_real_OrderedDict(self):
            d = OrderedDict(coll_OrderedDict(default_items))
            self.assertEqual(1, d['a'])

    def test_fromkeys(self):
        d = OrderedDict.fromkeys(['a', 'b', 'c'], 1)
        for k in ('a', 'b', 'c'):
            self.assertEqual(1, d[k])

    def test_brackets_mutator(self):
        d2 = OrderedDict()
        d2['d'] = 4
        self.assertEqual(d2['d'], 4)

    def test_del(self):
        d = OrderedDict(default_items)
        self.assertEqual(1, d['a'])
        del d['a']
        self.assertRaises(KeyError, lambda: d['a'])
    
    def test_setdefault_item_exists_nodefault(self):
        d = OrderedDict(default_items)
        self.assertEqual(1, d.setdefault('a'))
    def test_setdefault_item_exists_default(self):
        d = OrderedDict(default_items)
        self.assertEqual(1, d.setdefault('a', 4))
    def test_setdefault_item_not_exists_default(self):
        d = OrderedDict(default_items)
        self.assertEqual(4, d.setdefault('foo', 4))
        self.assertEqual(4, d['foo'])
    def test_setdefault_item_not_exists_nodefault(self):
        d = OrderedDict(default_items)
        self.assertEqual(None, d.setdefault('foo'))
        self.assertEqual(None, d['foo'])

    def test_clear(self):
        d = OrderedDict(default_items)
        d.clear()
        self.assertEqual(0, len(d))

    def test_pop_exists_nodefault(self):
        d = OrderedDict(default_items)
        self.assertEqual(1, d.pop('a'))
    def test_pop_exists_default(self):
        d = OrderedDict(default_items)
        self.assertEqual(1, d.pop('a', 12))
    def test_pop_not_exist_nodefault(self):
        d = OrderedDict(default_items)
        self.assertRaises(KeyError, lambda: d.pop('foo'))
    def test_pop_not_exist_default(self):
        d = OrderedDict(default_items)
        self.assertEqual(12, d.pop('foo', 12))
        
    def test_popitem_empty(self):
        d = OrderedDict()
        self.assertRaises(KeyError, lambda: d.popitem())
    def test_popitem_lifo(self):
        d = OrderedDict(default_items)
        self.assertEqual(('c', 3), d.popitem())
    def test_popitem_fifo(self):
        d = OrderedDict(default_items)
        self.assertEqual(('a', 1), d.popitem(last=False))

        
    def test_update_empty_source_with_odict(self):
        d1 = OrderedDict()
        d2 = OrderedDict(default_items)
        d1.update(d2)
        self.assertEqual(d1.items(), d2.items())
    if testing_real_ordered_dict:
        def test_update_empty_source_with_real_odict(self):
            d1 = OrderedDict()
            d2 = coll_OrderedDict(default_items)
            d1.update(d2)
            self.assertEqual(d1.items(), d2.items())
    def test_update_empty_source_with_list(self):
        d1 = OrderedDict()
        d2 = OrderedDict(default_items)
        d1.update(default_items)
        self.assertEqual(d1.items(), d2.items())
    def test_update_empty_source_with_dict(self):
        d1 = OrderedDict()
        d1.update(dict(default_items))
        d2 = OrderedDict(dict(default_items))
        self.assertEqual(d1.items(), d2.items())
    def test_update_empty_source_with_nothing(self):
        d = OrderedDict()
        d.update()
        self.assertEqual(0, len(d))

    def test_update_all_new_items(self):
        new_items = (('d', 4), ('e', 5))
        d1 = OrderedDict(default_items)
        d1.update(new_items)
        self.assertEqual(list(d1.items()),
                         list(default_items + new_items))
                  
    def test_update_some_old_items(self):
        new_items = (('a', 10), ('d', 4), ('e', 5))
        d1 = OrderedDict(default_items)
        d1.update(new_items)
        self.assertEqual(list(d1.items()),
                         [('a', 10), ('b', 2), ('c', 3), ('d', 4), ('e', 5)])
                  

# Obv. these would fail when testing the actual OrderedDictionary class...
if not testing_real_ordered_dict:    
    class TestNotImplemented(unittest.TestCase):
        def test_viewitems(self):
            self.assertRaises(NotImplementedError, 
                              lambda: OrderedDict().viewitems())
        def test_viewkeys(self):
            self.assertRaises(NotImplementedError, 
                              lambda: OrderedDict().viewkeys())
        def test_viewvalues(self):
            self.assertRaises(NotImplementedError, 
                              lambda: OrderedDict().viewvalues())
