#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@

from __future__ import division


import unittest

from rayon.backcompat import OrderedDict

import rayon.datatransformers
from rayon.datatransformers import *
from rayon.data import Dataset


# Usage based on rytimeseries.parse_data_relational
def usage_example_1():
    opts = {
        'group_by'       : 0, 
        'time_col'       : 1,
        'val_col'        : 2,
        'top_colname'    : 3,
        'btm_colname'    : 4,
        'top_lbl'        : "TOP",
        "btm_lbl"        : "BTM",
        'top_filter'     : lambda x: x <= 1000,
        'btm_filter'     : lambda x: x >= 0,
        'plot_label_map' : None
    }
    
    key_colnames = opts['group_by'].split(",")

    # row == a Row object from the incoming dataset
    # st  == state from the computation (e.g., previous path parts)
    # returns a bucket name in which to put the row
    def group_fn(rec, st):
        return [rec[x] for x in key_colnames]

    top_lbl = opts.get('top_lbl', default="TOP")
    passes_top = opts.get('top_filter', default=lambda x: True)
    btm_lbl = opts.get('btm_lbl', default="BOTTOM")
    passes_btm = opts.get('btm_filter', default=lambda x: False)


    # No bottom name implies a one-sided plot
    if opts['btm_colname']:
        # 2-sided function
        def side_fn(rec, st):
            if passes_top(rec['top_colname']):
                return top_lbl
            elif passes_top(rec['btm_colname']):
                return btm_lbl
    else:
        def side_fn(rec, st):
            if passes_top(rec['top_colname']):
                return top_lbl

            
    # args same as group_fn
    #
    # returns a tuple containing a projection of rec. (E.g., only the
    # time, value and key columns)
    def project_fn(rec, st):
        return tuple(rec[opts['time_col']],
                     rec[opts['val_col']])

    
    # three stages:
    #
    # 1. Partition the data by the keys in group_by. If plot_label_map
    #    contains an entry for a key, use that as the label;
    #    otherwise, concatenate the column names together and use
    #    that.
    #
    # 2. Filter out only the top and bottom side, (implicitly)
    #    discarding everything else. Use top/btm_label as label names,
    #    otherwise "TOP" and "BOTTOM".
    #
    # 3. Project the result into (time, value, (k1, k2, ...))
    #    tuples. (k1, etc. are columns in the key)
    #
    # The result is a tree. Leaf nodes contain an ordered collection
    # of records (no label). Interior nodes contain a label and a
    # collection of child nodes. The natural data structure for this
    # is an ordered map.
    xformer = (Classify(group_fn) + Classify(side_fn) + Project(project_fn))

    # results should be an OrderedDict whose keys are tuples of unique
    # key column values. Each value is another OrderedDict containing
    # one or two sides. The keys to the dict are the labels of the
    # side; the values are lists of (time, val, key) triples, where the
    # key is a tuple of all the key fields. E.g.:

    # results = OrderedDict(
    #     [('a', OrderedDict([('top', [(1,2,(3,4)),(4,5,(6,7))]),
    #                         ('btm', [(6,5,(4,3)),(5,4,(3,2))])])),
    #      ('b', ...)])


# ryscatterplot would be the same, without the second classify step.


"""
A tree node is either a Dataset or an OrderedDict whose keys are
hashable (duh) and whose values are tree nodes. IOW::


   TreeNode := Dataset a
             | OrderedDict(Hashable, TreeNode)

"""

# This dataset is used by most of the following tests:

d1_colnames = "x y z".split()
d1 = Dataset.from_rows([["a", 1, "alpha"],
                        ["a", 2, "beta"],
                        ["b", 1, "gamma"],
                        ["b", 2, "delta"]],
                       colnames=d1_colnames)

d1_nocolnames = Dataset.from_rows([["a", 1, "alpha"],
                                   ["a", 2, "beta"],
                                   ["b", 1, "gamma"],
                                   ["b", 2, "delta"]])
                                  


class t_lift(unittest.TestCase):

    new_dataset_1 = Dataset.from_rows(
        ["this is a new dataset".split()])
    new_dataset_2 = Dataset.from_rows(
        ["this is another new dataset".split()])


    def test_lift_1(self):
        def fn(dset, state):
            return self.new_dataset_1
        self.assertEqual(Lift(fn)(d1), self.new_dataset_1)
            

    def test_lift_2(self):
        "Test lift with composition."
        def lift_fn(dset, state):
            return self.new_dataset_1
        def group_fn(row):
            return [row[0]]
        self.assertEqual(
            (Classify(group_fn) + Lift(lift_fn))(d1),
            OrderedDict(
                [('a', self.new_dataset_1),
                 ('b', self.new_dataset_1)]))

    def test_lift_3(self):
        "Test ordering as a sequence"
        def fn(dset, state):
            return OrderedDict([('A', self.new_dataset_1), 
                                ('B', self.new_dataset_2)])
        ordering = ['B', 'A']

        self.assertEqual(
            Lift(fn, ordering)(d1), 
            OrderedDict([('B', self.new_dataset_2),
                         ('A', self.new_dataset_1)]))

    def test_lift_4(self):
        "Test ordering as an iterable"
        def fn(dset, state):
            return OrderedDict([('A', self.new_dataset_1), 
                                ('B', self.new_dataset_2)])
        ordering = (x for x in ['B', 'A'])

        self.assertEqual(
            Lift(fn, ordering)(d1), 
            OrderedDict([('B', self.new_dataset_2),
                         ('A', self.new_dataset_1)]))


    def test_lift_5(self):
        "Test ordering as a function returning a sequence"
        def fn(dset, state):
            return OrderedDict([('A', self.new_dataset_1), 
                                ('B', self.new_dataset_2)])
        def ordering(odict):
            return ['B', 'A']

        self.assertEqual(
            Lift(fn, ordering)(d1), 
            OrderedDict([('B', self.new_dataset_2),
                         ('A', self.new_dataset_1)]))

    def test_lift_6(self):
        "Test ordering as a function returning an iterable"
        def fn(dset, state):
            return OrderedDict([('A', self.new_dataset_1), 
                                ('B', self.new_dataset_2)])
        def ordering(odict):
            return (x for x in ['B', 'A'])

        self.assertEqual(
            Lift(fn, ordering)(d1), 
            OrderedDict([('B', self.new_dataset_2),
                         ('A', self.new_dataset_1)]))

    def test_lift_7(self):
        "Ordering ignores extra elements."
        def fn(dset, state):
            return OrderedDict([('A', self.new_dataset_1), 
                                ('B', self.new_dataset_2)])
        ordering = ['B', 'A', 'C']

        self.assertEqual(
            Lift(fn, ordering)(d1), 
            OrderedDict([('B', self.new_dataset_2),
                         ('A', self.new_dataset_1)]))

    def test_lift_8(self):
        "Path information is available to lifted functions."
        def fn1(dset, state):
            return OrderedDict([('A', dset)])
        def fn2(dset, state):
            parent = state['path'][-1]
            return OrderedDict([("%sB" % parent, dset)])
        def fn3(dset, state):
            parent = state['path'][-1]
            return OrderedDict([("%sC" % parent, dset)])
        self.assertEqual(
            (Lift(fn1) + Lift(fn2) + Lift(fn3))(d1),
            OrderedDict(
                [('A', OrderedDict(
                    [('AB', OrderedDict(
                        [('ABC', d1)]))]))]))
                        

        



class t_bind(unittest.TestCase):

    def test_bind_1(self):
        def group_1(row): return [row[0]]
        def group_2(row): return [row[1]]
        self.assertEqual(
            Bind(Classify(group_1), Classify(group_2))(d1),
            OrderedDict(
                [('a', OrderedDict([
                    (1, Dataset.from_rows([["a", 1, "alpha"]],
                                          colnames=d1_colnames)),
                    (2, Dataset.from_rows([["a", 2, "beta"]],
                                          colnames=d1_colnames))])),
                 ('b', OrderedDict([
                     (1, Dataset.from_rows([["b", 1, "gamma"]],
                                           colnames=d1_colnames)),
                     (2, Dataset.from_rows([["b", 2, "delta"]],
                                           colnames=d1_colnames))]))]))
                                        
    def test_bind_2(self):
        new_colnames = "x z".split()
        def group_1(row): return [row[0]]
        def group_2(row): return [row[1]]
        def project(row): return row[0], row[2]
        chain = (Classify(group_1) 
                 + Classify(group_2) 
                 + Project(project, colnames=new_colnames))
        self.assertEqual(
            chain(d1),
            OrderedDict(
                [('a', OrderedDict([
                    (1, Dataset.from_rows([["a", "alpha"]],
                                          colnames=new_colnames)),
                    (2, Dataset.from_rows([["a", "beta"]],
                                          colnames=new_colnames))])),
                 ('b', OrderedDict([
                     (1, Dataset.from_rows([["b", "gamma"]],
                                           colnames=new_colnames)),
                     (2, Dataset.from_rows([["b", "delta"]],
                                           colnames=new_colnames))]))]))

class t_classify(unittest.TestCase):
    def test_classify_1(self):
        def fn(row):
            if row[0] == 'a': return ["A"]
            else:             return ["B"]
        self.assertEqual(Classify(fn)(d1),
                         OrderedDict(
                             [('A', 
                               Dataset.from_rows(
                                   [["a", 1, "alpha"],
                                    ["a", 2, "beta"]],
                                   colnames=d1_colnames)),
                              ('B', 
                               Dataset.from_rows(
                                   [["b", 1, "gamma"],
                                    ["b", 2, "delta"]],
                                   colnames=d1_colnames))]))

    def test_classify_1_nocolnames(self):
        def fn(row):
            if row[0] == 'a': return ["A"]
            else:             return ["B"]
        self.assertEqual(Classify(fn)(d1_nocolnames),
                         OrderedDict(
                             [('A', 
                               Dataset.from_rows(
                                   [["a", 1, "alpha"],
                                    ["a", 2, "beta"]])),
                              ('B', 
                               Dataset.from_rows(
                                   [["b", 1, "gamma"],
                                    ["b", 2, "delta"]]))]))
        

    def test_classify_2(self):
        """
        Rows that are not classified (by returning None) are thrown
        away.
        """
        def fn(row):
            if row[0] == 'a' and row[1] == 1: return ["A"]
            elif row[0] == 'b' and row[1] == 1: return ["B"]

        self.assertEqual(Classify(fn)(d1),
                         OrderedDict(
                             [('A', 
                               Dataset.from_rows(
                                   [["a", 1, "alpha"]],
                                   colnames=d1_colnames)),
                              ('B', 
                               Dataset.from_rows(
                                   [["b", 1, "gamma"]],
                                   colnames=d1_colnames))]))

    def test_classify_3(self):
        """
        Ordering (as a sequence) is respected.
        """
        def fn(row):
            if row[0] == 'a' and row[1] == 1: return ["A"]
            elif row[0] == 'b' and row[1] == 1: return ["B"]

        self.assertEqual(Classify(fn, ['B', 'A'])(d1),
                         OrderedDict(
                             [('B', 
                               Dataset.from_rows(
                                   [["b", 1, "gamma"]],
                                   colnames=d1_colnames)),
                              ('A', 
                               Dataset.from_rows(
                                   [["a", 1, "alpha"]],
                                   colnames=d1_colnames))]))

    def test_classify_4(self):
        """
        Ordering (as an iterable) is respected.
        """
        def fn(row):
            if row[0] == 'a' and row[1] == 1: return ["A"]
            elif row[0] == 'b' and row[1] == 1: return ["B"]

        ordering = (x for x in ['B', 'A'])

        self.assertEqual(Classify(fn, ordering)(d1),
                         OrderedDict(
                             [('B', 
                               Dataset.from_rows(
                                   [["b", 1, "gamma"]],
                                   colnames=d1_colnames)),
                              ('A', 
                               Dataset.from_rows(
                                   [["a", 1, "alpha"]],
                                   colnames=d1_colnames))]))


    def test_classify_5(self):
        """
        Ordering (as a function returning a sequence) is respected.
        """
        def fn(row):
            if row[0] == 'a' and row[1] == 1: return ["A"]
            elif row[0] == 'b' and row[1] == 1: return ["B"]

        def ordering(odict):
            return ['B', 'A']

        self.assertEqual(Classify(fn, ordering)(d1),
                         OrderedDict(
                             [('B', 
                               Dataset.from_rows(
                                   [["b", 1, "gamma"]],
                                   colnames=d1_colnames)),
                              ('A', 
                               Dataset.from_rows(
                                   [["a", 1, "alpha"]],
                                   colnames=d1_colnames))]))

    def test_classify_6(self):
        """
        Ordering (as a function returning an iterable) is respected.
        """
        def fn(row):
            if row[0] == 'a' and row[1] == 1: return ["A"]
            elif row[0] == 'b' and row[1] == 1: return ["B"]

        def ordering(odict):
            return (x for x in ['B', 'A'])

        self.assertEqual(Classify(fn, ordering)(d1),
                         OrderedDict(
                             [('B', 
                               Dataset.from_rows(
                                   [["b", 1, "gamma"]],
                                   colnames=d1_colnames)),
                              ('A', 
                               Dataset.from_rows(
                                   [["a", 1, "alpha"]],
                                   colnames=d1_colnames))]))

    def test_classify_7(self):
        "StClassify passes state into classifier."
        def fn(row, state):
            if row[0] == 'a': return ["A"]
            else:             return ["B"]
        self.assertEqual(StClassify(fn)(d1),
                         OrderedDict(
                             [('A', 
                               Dataset.from_rows(
                                   [["a", 1, "alpha"],
                                    ["a", 2, "beta"]],
                                   colnames=d1_colnames)),
                              ('B', 
                               Dataset.from_rows(
                                   [["b", 1, "gamma"],
                                    ["b", 2, "delta"]],

                                   colnames=d1_colnames))]))

    def test_classify_7_nocolnames(self):
        "StClassify passes state into classifier."
        def fn(row, state):
            if row[0] == 'a': return ["A"]
            else:             return ["B"]
        self.assertEqual(StClassify(fn)(d1_nocolnames),
                         OrderedDict(
                             [('A', 
                               Dataset.from_rows(
                                   [["a", 1, "alpha"],
                                    ["a", 2, "beta"]])),
                              ('B', 
                               Dataset.from_rows(
                                   [["b", 1, "gamma"],
                                    ["b", 2, "delta"]]))]))
    

    def test_classify_8(self):
        "Rows can be classified into multiple groups."
        def fn(row):
            if row[0] == 'a': return ["A", "altA", "all"]
            else:             return ["B", "altB", "all"]
        self.assertEqual(Classify(fn)(d1),
                         OrderedDict(
                             [('A', 
                               Dataset.from_rows(
                                   [["a", 1, "alpha"],
                                    ["a", 2, "beta"]],
                                   colnames=d1_colnames)),
                              ('altA', 
                               Dataset.from_rows(
                                   [["a", 1, "alpha"],
                                    ["a", 2, "beta"]],
                                   colnames=d1_colnames)),
                              ('all', 
                               Dataset.from_rows(
                                   [["a", 1, "alpha"],
                                    ["a", 2, "beta"],
                                    ["b", 1, "gamma"],
                                    ["b", 2, "delta"]],
                                   colnames=d1_colnames)),
                              ('B', 
                               Dataset.from_rows(
                                   [["b", 1, "gamma"],
                                    ["b", 2, "delta"]],
                                   colnames=d1_colnames)),
                              ('altB', 
                               Dataset.from_rows(
                                   [["b", 1, "gamma"],
                                    ["b", 2, "delta"]],
                                   colnames=d1_colnames))]))

class t_transformers(unittest.TestCase):                

    def test_project_1(self):
        def fn(row):
            return row[0], row[2]
        self.assertEqual(Project(fn, colnames="aa bb".split())(d1),
                         Dataset.from_rows(
                             [["a", "alpha"],
                              ["a", "beta"],
                              ["b", "gamma"],
                              ["b", "delta"]],
                             colnames="aa bb".split()))


    def test_project_2(self):
        def fn(row):
            return row[0], (row[2] + "a")
        self.assertEqual(Project(fn, colnames="aa bb".split())(d1),
                         Dataset.from_rows(
                             [["a", "alphaa"],
                              ["a", "betaa"],
                              ["b", "gammaa"],
                              ["b", "deltaa"]],
                             colnames="aa bb".split()))

    def test_reorder_1(self):
        self.assertEqual(Reorder(new_colnames="x z".split())(d1),
                         Dataset.from_rows(
                             [["a", "alpha"],
                              ["a", "beta"],
                              ["b", "gamma"],
                              ["b", "delta"]],
                             colnames="x z".split()))

    def test_reorder_2(self):
        "Reorder can take a function that returns column names."
        def fn(dset):
            return "x z".split()
        self.assertEqual(Reorder(new_colnames=fn)(d1),
                         Dataset.from_rows(
                             [["a", "alpha"],
                              ["a", "beta"],
                              ["b", "gamma"],
                              ["b", "delta"]],
                             colnames="x z".split()))

    def test_reorder_3(self):
        "StReorder can take a function that takes state."
        def fn(dset, state):
            return "x z".split()
        self.assertEqual(StReorder(new_colnames=fn)(d1),
                         Dataset.from_rows(
                             [["a", "alpha"],
                              ["a", "beta"],
                              ["b", "gamma"],
                              ["b", "delta"]],
                             colnames="x z".split()))



    def test_include_1(self):
        def fn(row):
            return row[0] == "a"
        self.assertEqual(Include(fn)(d1),
                         Dataset.from_rows(
                             [["a", 1, "alpha"],
                              ["a", 2, "beta"]],
                             colnames=d1_colnames))

    def test_include_1_nocolnames(self):
        def fn(row):
            return row[0] == "a"
        self.assertEqual(Include(fn)(d1_nocolnames),
                         Dataset.from_rows(
                             [["a", 1, "alpha"],
                              ["a", 2, "beta"]]))

    def test_include_2(self):
        "StInclude passes state to the predicate."
        def fn(row, st):
            return row[0] == "a"
        self.assertEqual(StInclude(fn)(d1),
                         Dataset.from_rows(
                             [["a", 1, "alpha"],
                              ["a", 2, "beta"]],
                             colnames=d1_colnames))
        

    def test_exclude_1(self):
        def fn(row):
            return row[0] == "a"
        self.assertEqual(Exclude(fn)(d1),
                         Dataset.from_rows(
                             [["b", 1, "gamma"],
                              ["b", 2, "delta"]],
                             colnames=d1_colnames))

    def test_exclude_1_nocolnames(self):
        def fn(row):
            return row[0] == "a"
        self.assertEqual(Exclude(fn)(d1_nocolnames),
                         Dataset.from_rows(
                             [["b", 1, "gamma"],
                              ["b", 2, "delta"]]))


    def test_exclude_2(self):
        "StExclude passes state to the predicate."
        def fn(row):
            return row[0] == "a"
        self.assertEqual(StExclude(fn)(d1),
                         Dataset.from_rows(
                             [["b", 1, "gamma"],
                              ["b", 2, "delta"]],
                             colnames=d1_colnames))


