#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@

from __future__ import division

"""
``testutils`` --- utility functions for testing
"""


class _Mock(object):
    def __init__(self):
        self._calls = []
    def check_call(self, call_id, name, *args, **kargs):
        try:
            (c_name, c_args, c_kargs) = self._calls[call_id]
        except IndexError:
            raise AssertionError("no call #%d (expected '%s')" % (
                    call_id, name))
        if c_name != name:
            raise AssertionError("check_calls failed: name '%s' != '%s'" %
                                 (c_name, name))
        if c_args != args:
            raise AssertionError("check_calls failed: args %s != %s" %
                                 (c_args, args))
        if c_kargs != kargs:
            raise AssertionError("check_calls failed: keyword args %s != %s" %
                                 (c_kargs, kargs))

class _Returner(object):
    # Strictly a typed function
    def __init__(self, fn): self._fn = fn
    def __call__(self, *args, **kargs):
        return self._fn(*args, **kargs)

class _BoundMethod(object):
    # Contains a method that should be bound to an object
    # called
    def __init__(self, fn): self._fn = fn
    def bind(self, obj, cls):
        return self._fn.__get__(obj, cls)
    
def Raises(e):
    """
    Indicates that a function call shoud raise *e*. 
    """
    def fn(*a, **ka): raise e
    return _Returner(fn)

def Cycle(*rvs):
    """
    Indicates that a function should return *rvs* in order with each
    successive call. *rvs* may contain the special values
    :func:`Raises` or (though it's probably unnecessary) :func:`Cycle`.
    """
    def gen():
        while True:
            for r in rvs:
                yield r
    g = gen()
    def fn(*a, **ka):
        rc = g.next()
        if isinstance(rc, _Returner):
            return rc(*a, **ka)
        else:
            return rc
    return _Returner(fn)

def BoundMethod(fn):
    return _BoundMethod(fn)


def Mock(method_dict=None):
    """
    Returns a mock object. *method_dict* is a dictionary whose keys
    are method calls. By default, this method will return the value
    associated with the key in *method_dict* when called. Some special
    values change behavior:

       * If value is the output of the :func:`Raises` function, an
         exception will be raised.

       * If the value is the output of the :func:`Cycle` function,
         the method will return different values for each call. If the
         method is called more times than values are specified, it
         will return the value from the beginning of the sequence and
         continue from there.
    """
    mockery = _Mock()
    if method_dict is not None:
        def mk_method(meth_name, return_val):
            def method(*args, **kargs):
                mockery._calls.append((meth_name, args, kargs))
                if isinstance(return_val, _Returner):
                    return return_val()
                else:
                    return return_val
            return method
            
            
        for k, v in method_dict.items():
            # special case for bound methods
            if isinstance(v, _BoundMethod):
                setattr(mockery, k, v.bind(mockery, type(mockery)))
            else:
                setattr(mockery, k, mk_method(k, v))
    return mockery


__all__ = """
BoundMethod
Cycle
Mock
Raises
""".split()
