#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
from __future__ import division
import math
import copy
import re

import markers
import lines
import miscutils

from rayon import RayonException

font_base_size = 10

point_size_thresholds = {
    'small'     : ( 8, 9),
    'normal'    : ( 9, 10, 12),
    'large'     : (14, 18, 24),
    'x-large'   : (28, 32, 34, 36),
    'xx-large'  : (40, 42, 44, 46, 48)
}

font_size_modifiers = {
    'small'    : 0.5,  # 6px (but we threshold at 8)
    'normal'   : 1.00, # 12 px, e.g.
    'large'    : 2.00, # 24 px
    'x-large'  : 3.00, # 36 px
    'xx-large' : 4.00, # 48 px
}

font_size_names = set(font_size_modifiers.keys())

def get_point_size_from_relative(font_size, canvas_width, canvas_height):
    """Map a relative font size to a point size, given the width and
    height of the whole canvas."""
    scale_factor = (min(canvas_width  / 1400, 1) +
                    min(canvas_height / 1400, 1))
    raw_point_size = max(point_size_thresholds[font_size][0], 
                         font_base_size
                         * font_size_modifiers[font_size]
                         * scale_factor)
    point_size = point_size_thresholds[font_size][-1]
    for idx, sz in enumerate(point_size_thresholds[font_size]):
        if raw_point_size > sz:
            continue
        if raw_point_size == sz:
            point_size = raw_point_size
            break
        last_sz = point_size_thresholds[font_size][idx - 1]
        if ((sz - raw_point_size) < (raw_point_size - last_sz)):
            point_size = sz
        else:
            point_size = last_sz
        break

    #print "(%s -> %s)" % (raw_point_size, point_size)
    return point_size

def get_point_size(font_size, canvas_width, canvas_height):
    """Return a size given a font size specification.

    Decimal sizes will be rounded down to the smallest whole
    number. Negative numbers are converted to their absolute value.

    This function should really be named get_size, since the returned
    value will render in canvas units, which is points in PDF and
    pixels everywhere else.
    """
    # If it's a number, give it back
    if isinstance(font_size, (int, float)):
        return abs(int(font_size))

    if isinstance(font_size, basestring):
        # If it's a relative font spec, convert it
        try:
            if font_size in font_size_names:
                # Return the "appropriate" font size
                return get_point_size_from_relative(
                    font_size, canvas_width, canvas_height)
        except KeyError:
            raise RayonException("Unknown font specification: '%s'" % font_size)

        # If it's a numeric string (with our without units), convert
        # it
        if font_size.endswith("px") or font_size.endswith("pt"):
            font_size = font_size[:-2]
        try:
            return abs(int(font_size))
        except ValueError:
            try:
                return abs(int(float(font_size)))
            except ValueError:
                # It isn't convertible to int -- fall through
                pass
                    
    # Don't know what it is
    raise RayonException("Unknown font specification: '%s'" % font_size)

# TODO: fix prism so this isn't necessary
import old_font_stuff
def use_old_get_point_size():
    global get_point_size
    get_point_size = old_font_stuff.get_point_size
    

def parse_borders(all, left, right, top, bottom):
    # return (left, right, top, bottom)
    if all is not None:
        all = lines.create_from_spec(all)
    rc = []
    for b in left, right, top, bottom:
        if b is None:
            if all is None:
                rc.append(None)
            else:
                rc.append(all)
        else:
            rc.append(lines.create_from_spec(b))
    return rc



def get_bbox(content_width, content_height, halign, valign, angle=0):
    """Get the bounding box for a label of a given width and height,
    adjusted for the given horizontal and vertical alignment and
    angle.

    Coordinates are given as offsets from the origin. For a label with
    a horizontal alignment of C{left} and a vertical alignment of
    C{bottom}, the origin represents the top left corner. For a
    horizontal/vertical alignment of C{center/center}, it represents
    the center of the content box. For c{bottom/right}, it represents
    the right corner of the box.

    @type  content_width: int
    @param content_width: width of label contents

    @type  content_height: int
    @param content_height: height of label contents

    @type  halign: string
    @param halign: one of the string literals C{left}, C{center} or
        C{right}, indicating the horizontal alignment of the contents.

    @type  valign: string
    @param valign: one of the string literals C{top}, C{center} or
        C{bottom}, indicating the vertical alignment of the contents.

    @type  angle: float
    @param angle: degrees of rotation, counterclockwise from
        horizontal, rotating around C{0,0}.

    @rtype: 4-tuple
    @return: Four C{(x,y)} coordinate pairs respectively
        representing the top left, bottom left, top right and bottom
        right corners of the bounding box for the component.
    """
    # Get basic offsets
    if halign == "left":     offset_x = 0
    elif halign == "center": offset_x = content_width / 2
    elif halign == "right":  offset_x = content_width

    if valign == "top":      offset_y = 0
    elif valign == "center": offset_y = content_height / 2
    elif valign == "bottom": offset_y = content_height

    # we'll need these a lot, so...
    inv_offset_x = content_width - offset_x
    inv_offset_y = content_height - offset_y

    # For reference on the rest of this, consult
    # etc/rectangle-offset-diagram.pdf and
    # etc/rectangle-offset-diagram.org (which shows the work behind
    # this in more detail). The comments below will refer to the two
    # diagrams in this file.


    if angle == 0:
        # No angle. Easy.
        topleft = -offset_x, -offset_y
        btmleft = -offset_x, inv_offset_y
        topright = inv_offset_x, -offset_y
        btmright = inv_offset_x, inv_offset_y
    else:
        # Not so easy. First, normalize angle so it isn't negative
        if angle < 0:
            angle = 360 + angle

        angle = math.radians(angle)

        # Functions to calculate the polar coordinates of points A, B,
        # C and D with respect to E
        def e2a():
            r = math.sqrt((inv_offset_x ** 2) + (offset_y ** 2))
            theta = math.asin(offset_y / r) + angle
            return r, theta
        def e2b():
            r = math.sqrt((offset_x ** 2) + (offset_y ** 2))
            theta = math.asin(offset_x / r) + math.radians(90) + angle
            return r, theta
        def e2c():
            r = math.sqrt((offset_x ** 2) + (inv_offset_y ** 2))
            theta = (math.asin(inv_offset_y / r) + math.radians(180) + angle)
            return r, theta
        def e2d():
            r = math.sqrt((inv_offset_x ** 2) + (inv_offset_y ** 2))
            theta = (math.asin(inv_offset_x / r) + math.radians(270) + angle)
            return r, theta
            
        if offset_x == 0: # C
            if offset_y == 0:
                # No offsets. (3C)
                r_a, theta_a = inv_offset_x, angle
                r_b, theta_b = 0, 0
                r_c, theta_c = inv_offset_y, math.radians(270) + angle
                r_d, theta_d = e2d()
            elif inv_offset_y == 0:
                # 1C
                r_a, theta_a = e2a()
                r_b, theta_b = offset_y, math.radians(90) + angle
                r_c, theta_c = 0, 0
                r_d, theta_d = inv_offset_x, angle
            else:
                # 2C
                r_a, theta_a = e2a()
                r_b, theta_b = offset_y, math.radians(90) + angle
                r_c, theta_c = inv_offset_y, math.radians(270) + angle
                r_d, theta_d = e2d()
        elif inv_offset_x == 0: # A
            if offset_y == 0:
                # 3A
                r_a, theta_a = 0, 0
                r_b, theta_b = offset_x, math.radians(180) + angle
                r_c, theta_c = e2c()
                r_d, theta_d = inv_offset_y, math.radians(270) + angle
            elif inv_offset_y == 0:
                # 1A
                r_a, theta_a = offset_y, math.radians(90) + angle
                r_b, theta_b = e2b()
                r_c, theta_c = offset_x, math.radians(180) + angle
                r_d, theta_d = 0, 0
            else:
                # 2A
                r_a, theta_a = offset_y, math.radians(90) + angle
                r_b, theta_b = e2b()
                r_c, theta_c = e2c()
                r_d, theta_d = inv_offset_y, math.radians(270) + angle
        else: # B
            # note: 0 < offset_x < content_width
            if offset_y == 0:
                # 3B
                r_a, theta_a = inv_offset_x, angle
                r_b, theta_b = offset_x, math.radians(180) + angle
                r_c, theta_c = e2c()
                r_d, theta_d = e2d()
            elif inv_offset_y == 0:
                # 1B
                r_a, theta_a = e2a()
                r_b, theta_b = e2b()
                r_c, theta_c = offset_x, math.radians(180) + angle
                r_d, theta_d = inv_offset_x, angle
            else:
                # 2B
                r_a, theta_a = e2a()
                r_b, theta_b = e2b()
                r_c, theta_c = e2c()
                r_d, theta_d = e2d()
                
        # In our coordinate system, Y runs down, not up, so we have to
        # invert it.
        topright = (r_a * math.cos(theta_a), -(r_a * math.sin(theta_a)))
        topleft  = (r_b * math.cos(theta_b), -(r_b * math.sin(theta_b)))
        btmleft  = (r_c * math.cos(theta_c), -(r_c * math.sin(theta_c)))
        btmright = (r_d * math.cos(theta_d), -(r_d * math.sin(theta_d)))


    return topleft, btmleft, topright, btmright

        
class Labeler(object):
    """
    Draws text labels of a given style and alignment to parts of
    a chart.

    *font_size*, *font_family* and *font_style* define the font used
    for the label, and take the values described in
    :ref:`specifying-text-properties`. *halign* and *valign* specify
    the horizontal and vertical alignment of the label.

    *color* is the color of the label text. *bgcolor* is the color of
    the label background. *angle* is the angle of the text (0 ==
    horizontal, left to right).

    If specified, *border* is a text line specification (see
    :ref:`specifying-lines`) describing the :class:`~rayon.lines.Line`
    object to use to draw border lines around the label text's
    bounding box. *border_left*, *border_right*, *border_top* and
    *border_bottom* specify lines that will only be drawn on the left,
    right, top and bottom edges of the box, respectively. If *border*
    is specified, the other *border_* arguments will be ignored.
    """
    def __init__(
        self, font_size="normal", font_family="default",
        font_style="normal", font_weight="normal",
        halign="left", valign="center",
        color=(0, 0, 0), bgcolor=None, angle=0,
        border=None, border_top=None, border_bottom=None,
        border_left=None, border_right=None):
        # TODO: verify that these font metrics exist....
        self.font_size = font_size
        self.font_family = font_family
        self.font_style = font_style
        self.font_weight = font_weight
        if halign in ("left", "center", "right"):
            self.halign = halign
        else:
            raise RuntimeError("Invalid value for halign: %s" % halign)
        if valign in ("top", "center", "bottom"):
            self.valign = valign
        else:
            raise RuntimeError("Invalid value for valign: %s" % valign)
        self.color = miscutils.parse_color(color)
        self.angle = angle
        self.borders = parse_borders(border,
                                     border_left, border_right,
                                     border_top, border_bottom)
        if bgcolor is not None:
            self.bgcolor = miscutils.parse_color(bgcolor)
        else:
            self.bgcolor = None

    def get_horizontal_alignment(self): return self.halign
    def get_vertical_alignment(self): return self.valign 

    def __mul__(self, num):
        return [copy.copy(self) for _ in xrange(num)]

    def get_bounding_box(self, ctx, content):
        """Get the bounding box for a label containing I{content}.

        @type  ctx: C{backend_*.Context}
        @param ctx: backend-specific context for rendering the label

        @type  content: string
        @param content: content of the label.

        @rtype: tuple
        
        @return: the 5-tuple C{(topleft, btmleft, topright, btmright,
            point_size)} containing, respectively, C{(x,y)}
            coordinates of the top left, bottom, left, top right, and
            bottom right corners of the bounding box, and the
            calculated point size of the type."""
        canvas_width, canvas_height = ctx.canvas_dimensions()
        pointsize = get_point_size(self.font_size, canvas_width, canvas_height)

        ex_w, ex_h = ctx.get_text_dimensions(
            content, pointsize, self.font_family, self.font_style,
            self.font_weight)

        corners = get_bbox(ex_w, ex_h, self.halign, self.valign, self.angle)
        return corners + (pointsize,)

    def draw(self, ctx, x, y, width, height, content):
        """ Draw label. If the label has a horizontal alignment of
        C{left} and a vertical alignment of C{bottom}, I{x} and I{y}
        refer to the upper left corner of the label box.

        @type ctx: C{backend_*.Context}
        @param ctx: context into which to draw
        
        @type x: int
        @param x: x-position of label.
        
        @type y: int
        @param y: y-position of label

        @type  width: number
        @param width: width of enclosing component object

        @type  height: number
        @param height: height of enclosing component object

        @type  content: string
        @param content: text content of label
        """
        # Don't bother if there's nothing to draw
        if content == "" or content is None:
            return
        
        # width and height are currently ignored. They are mostly here
        # to provide a uniform interface with
        # ColumnLabeler.draw(). (They could be used to clip to a
        # bounding box at some point. Currently, this is done at a
        # higher level. See plots.LabeledScatterPlot and
        # markers.LabeledMarker for examples.)
        
        (topleft,  btmleft,
         topright, btmright, pointsize) = self.get_bounding_box(ctx, content)

        # Gack. TODO: standardize bounding boxes better. topleft,
        # bottomleft, etc. doesn't naturally draw a box, it draws a
        # bowtie.

        bbox = tuple((x + i[0], y + i[1])
                     for i in (topleft, btmleft, topright, btmright))

        poly_bbox = tuple((x + i[0], y + i[1])
                     for i in (topleft, btmleft, btmright, topright))

        # Draw background if necessary
        # xXx TEST
        # ctx.draw_polygon(poly_bbox, 0, (0, 0, 0, 0), (0, 0, 0, 0), True)
        
        if self.bgcolor is not None:
            ctx.draw_polygon(poly_bbox, 0, (0, 0, 0, 0), self.bgcolor, True)

                             
        ctx.draw_text(content, bbox,
                      pointsize, self.font_family, self.font_style,
                      self.font_weight, self.color, self.angle)
        
        # Draw borders as necessary
        left, right, top, bottom = self.borders
        for bline, _from, _to in ((left, topleft, btmleft),
                            (right, topright, btmright),
                            (top, topleft, topright), 
                           (bottom, btmleft, btmright)):
            if bline is not None:
                bline.draw(ctx,
                           x + _from[0], y + _from[1],
                           x + _to[0], y + _to[1])

        
class StaticText(Labeler):
    "A L{Labeler} with static textual content."
    def __init__(self, text, *args, **kwargs):
        Labeler.__init__(self, *args, **kwargs)
        self.content = text
    def get_bounding_box(self, ctx, content=None):
        # NB: For compatibility with the superclass interface, we take
        # the content argument (optionally) but ignore it.
        return Labeler.get_bounding_box(self, ctx, self.content)
    def draw(self, ctx, x, y, width, height, content=None):
        # NB: For compatibility with the superclass interface, we take
        # the content argument (optionally) but ignore it.
        return Labeler.draw(self, ctx, x, y, width, height, self.content)




class ColumnLabeler(object):
    def __init__(self, labelers, box_width, col_width="auto",
                 halign="left", valign="center",
                 gutter=0,
                 border=None, border_top=None, border_bottom=None,
                 border_left=None, border_right=None):
        """Constructor.

        @type  box_width: float between 0 and 1
        @param box_width: width of total label, as a proportion of the
            enclosing component (usually a border)

        @param  col_width: one of
            - The string literal C{equal}. All columns will have equal
              width
            - A sequence of template strings. There must be as many
              template strings as there are columns. Columns will be
              wide enough to accomodate the template strings.
            - A sequence of numbers between 0 and 1, representing a
              percentage of the total available width. There must be
              as many numbers as there are columns.
            - A sequence of functions that will be called during
              drawing. Each function I{f} takes the form I{f(ctx,
              width) -> col_width}, where I{ctx} is the backend
              drawing context, I{width} is the width of the
              C{ColumnLabeler}, and I{col_width} is the width of the
              column.
        """
        self.labelers = labelers
        self.box_width = box_width
        self.col_width = col_width
        self.halign = halign
        self.valign = valign
        self.gutter = gutter
        self.borders = parse_borders(border,
                                     border_left, border_right,
                                     border_top, border_bottom)


    def get_horizontal_alignment(self): return self.halign
    def get_vertical_alignment(self): return self.valign

    def __mul__(self, num):
        return [copy.copy(self) for _ in xrange(num)]

    def draw(self, ctx, x, y, width, height, content):
        # content is presumed to be a tuple of strings
        assert(len(content) == len(self.labelers))



        # Box width, column width, etc.
        scaled_width = width * self.box_width
        if self.col_width == "auto":
            col_width = []
            for idx, l in enumerate(self.labelers):
                bbox = l.get_bounding_box(ctx, content[idx])
                col_width.append(abs(bbox[0][0] - bbox[2][0]))
        elif self.col_width == "equal":
            col_width = ([scaled_width / len(self.labelers),]
                         * len(self.labelers))
        elif (not isinstance(self.col_width, str) and
              len(self.col_width) == len(self.labelers) and
              all(isinstance(_, str) for _ in self.col_width)):
            # It's a template string
            col_width = []
            for l, template in zip(self.labelers, self.col_width):
                bbox = l.get_bounding_box(ctx, template)
                col_width.append(abs(bbox[0][0] - bbox[2][0]))
        elif (all(callable(cw) for cw in self.col_width)):
            # It's a list of functions that compute the column width
            col_width = [cw(ctx, scaled_width) for cw in self.col_width]
        else:
            # It's a list of numbers, or it should be
            try:
                # Verify that they're numbers
                for _ in self.col_width:
                    _ + 1
            except:
                raise RuntimeError("All elements in numeric column width "
                                   "specifier should be numeric.")
            # Verify that they sum to <= 1
            if sum(self.col_width) > 1:
                raise RuntimeError("Total width specification exceeds 1")
            col_width = [scaled_width * cw for cw in self.col_width]


        # TODO: At this point, the total width of all the columns can
        # exceed the width of the box. I'm currently just overflowing
        # the box here, but some other options are to clip the text to
        # fit or generate point sizes from a different set of
        # heuristics that assumes the font can fit more of the width
        # (then compute point size using the border box width instead
        # of the canvas width).
        #
        # No matter what, bad things can happen if the data and the
        # formatting don't play nicely. That last one probably fails
        # most gracefully, but can still produce oddly sized or
        # unreadable type.
        

        # use the total width of all labels, plus gutters (unless it
        # exceeds the total scaled width)
        gutter_width = self.gutter * len(col_width)
        label_width = min(scaled_width, sum(col_width) + gutter_width)
        # use the height of the tallest label in the row
        label_height = max(
            l.get_bounding_box(ctx, content[idx])[1][1]
            for idx, l in enumerate(self.labelers))

        (topleft,  btmleft,
         topright, btmright) = get_bbox(label_width, label_height,
                                        self.halign, self.valign)


        # I'm assuming that the underlying labelers have a valign of
        # "top." If they all have the same non-"top" value, the whole
        # box will look as if it's shifted up. If they're different
        # from each other, we're in Krazy Klown Land. So users, if
        # you're reading this, don't do that.

        real_topleft = (topleft[0] + x, topleft[1] + y)
        offset = 0
        for idx, label in enumerate(self.labelers):
            # Advance X past half the gutter
            offset += self.gutter/2
            halign_offset=0
            label_halign = label.get_horizontal_alignment()
            # Adjust X for horizontal alignment of underlying label
            if label_halign == "center":
                halign_offset = (col_width[idx] / 2)
            elif label_halign == "right":
                halign_offset = col_width[idx]
            elif label_halign == "left":
                # No action, it's fine where it is
                pass
            offset += halign_offset
            # Do the draw (TODO: perhaps width should be scaled_width....)
            label.draw(ctx, real_topleft[0] + offset, y, width, height,
                       content[idx])
            # Advance X past this column (and the other half of the
            # gutter)
            offset += col_width[idx] + (self.gutter / 2) - halign_offset

        
        # Draw borders as necessary
        left, right, top, bottom = self.borders
        for bline, _from, _to in ((left, topleft, btmleft),
                            (right, topright, btmright),
                            (top, topleft, topright),
                            (bottom, btmleft, btmright)):
            if bline is not None:
                bline.draw(ctx,
                           x + _from[0], y + _from[1],
                           x + _to[0], y + _to[1])


    
# TODO: a RowLabeler to label X axes in the analogous way.

