#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
from __future__ import division

"""A factory object for the various tools used to assemble a Rayon chart."""

import sys

from rayon import RayonException
import backend
import borders
import charts
import data
import grids
import lines
import markers
import miscutils
import plots
import scales
import text
import tickmarkers
import ticks

class UnsupportedBackend(RayonException):
    def __str__(self):
        return "Unsupported backend: %s" % RayonException.__str__(self)
    def backend(self): return RayonException.__str__(self)

class UnsupportedOperationForBackend(RayonException):
    def __init__(self, backend, op):
        self.backend_name = backend.name
        self.op = op
    def backend(self): return self.backend_name
    def __str__(self):
        return "Backend '%s' cannot %s" % (backend, op)

# Supported backends (we always have the test backend)
import backend_test_static
available_backends = { 'test_static' : backend_test_static }

# Do we have cairo
import backend_cairo
if backend_cairo.is_available():
    available_backends[backend_cairo.name] = backend_cairo


# Do we have WX?
import backend_wx
if backend_wx.is_available():
    available_backends[backend_wx.name] = backend_wx
    



class Toolbox(object):
    """
    A factory object for creating most of the objects available in
    Rayon. See :ref:`toolbox` for more information.

    :class:`Toolbox` objects should be created with one of the
    :meth:`Toolbox.for_file` or :meth:`Toolbox.for_gui` methods,
    which will properly associate the :class:`Toolbox` object with a
    rendering backend.
    """

    # Factories
    
    @classmethod
    def for_backend(cls, backend_name):
        """
        Create a :class:`Toolbox` object for the rendering backend
        *backend_name*. Valid names are ``cairo`` and ``wx``. (Both
        backends may not be available in your installation. Consult
        your system administrator to determine what backends are
        available on your system.)
        """ 
        try:
            backend = available_backends[backend_name]
        except KeyError, e:
            raise UnsupportedBackend(backend_name)
        return cls(backend)
    
    @classmethod
    def for_file(cls):
        """
        Creates a :class:`Toolbox` object appropriate for rendering
        static visualizations to a file.
        """
        return cls.for_backend('cairo')

    @classmethod
    def for_gui(cls, gui="wx"):
        """
        Create a :class:`Toolbox` object with a backend for rendering
        to a GUI.

        *gui* is a string label identifying the GUI backend to use. At
        this time, specifying any value except ``wx`` will raise an
        exception.
        """
        if gui != wx:
            raise UnsupportedBackend(gui)
        return cls.for_backend('wx')

    @classmethod
    def for_test(cls, ttype='static'):
        """
        Create a toolbox for running in a test context.
        """
        # Undocumented method
        if ttype != 'static':
            raise UnsupportedBackend(ttype)
        return cls.for_backend('test_static')
            

    def __init__(self, backend):
        self.backend = backend


        
    def get_backend(self):
        return self.backend

    #
    # Pages
    #

    def new_page_from_buffer(self, buff):
        """
        Create a :ref:`page <ref-page>` that can be used to
        render a :ref:`chart <ref-charts>` into *buff*. *buff*
        should be an in-memory buffer object from the system
        associated with the rendering backend, representing a
        2-dimensional pixel surface.

        The rendering backend must support rendering to memory.
        """
        if not self.backend.output_to_memory:
            return UnsupportedOperationForBackend(
                self.backend, "write to memory")
        return self.backend.Page.from_buffer(buff)

    new_page_from_buffer.meta = (backend.Page, None)
        
    def new_page_from_filename(self, filename, width, height):
        """
        Create a :ref:`page <ref-page>` that can be used to render a
        :ref:`chart <ref-charts>` to *filename* with dimensions
        *width* and *height*. The type of file will be selected based
        on file extension. Currently supported file formats are SVG,
        PDF and PNG.
        """
        if not self.backend.output_to_file:
            return UnsupportedOperationForBackend(
                self.backend, "write to a file by name")
        return self.backend.Page.from_filename(filename, width, height)

    new_page_from_filename.meta = (backend.Page, None)

    def new_page_from_logger(self, logger=lambda a, b, c: None,
                             width=800, height=400):
        """
        Create a :ref:`page <ref-page>` that can be used to render a
        :ref:`chart <ref-charts>` to a command logger. This is used to
        test Rayon functionality.

        *logger* is a function that takes a string (representing a
        method call on a backen context object) and a dictionary (the
        arguments to that method call) and returns nothing. It
        presumably does something to note the method call.
        """
        # This method's documentation is not published.
        return self.backend.Page.from_logger(logger, width, height)

    def new_dataset_from_stream(self, stream, colnames=None,
                                typemap=None, delimiter="|",
                                first_line_is_colnames=False):
        """
        Creates a :ry:factory:`~rayon.data.Dataset` from data read
        from an input stream. The format of the streamed data is
        described in :ref:`On-disk format <data-on-disk-format>`.

        If supplied, *colnames* is a sequence of strings containing
        column names for the data. There must be as many names in the
        sequence as there are columns in the data. If not specified,
        column names from the file will be used; if the file contains
        no column names, the :ry:factory:`~rayon.data.Dataset` will be
        created without column names.

        If supplied, *typemap* is a list of functions that take a
        string representation of an object and return the
        object. There must be as many functions in the sequence as
        there are columns in the data. If not specified, strings
        composed entirely of numbers (optionally with a single period)
        will be converted to numbers, and everything else will be left
        as strings.

        If supplied, *delimiter* is the character used in the file to
        delimit fields in a record. The default delimiter is the pipe
        (``|``) character. The delimiter may also be specified in the
        file format; in that case, this argument is ignored.

        If *first_line_is_colnames* is ``True`` and no other column
        names are supplied in the file (i.e., with a header or a
        specially-commented first line), assume the first line of the
        input is a delimited list of column names. Use this when
        parsing input from external programs (such as the SiLK tools)
        which pass their column names as the first line in a file but
        do not indicate them in any other way.
        """
        return data.Dataset.from_stream(
            stream, colnames, typemap, delimiter, first_line_is_colnames)

    new_dataset_from_stream.meta = (data.Dataset, None)

    def new_dataset_from_filename(self, filename, colnames=None,
                                  typemap=None, delimiter="|",
                                  first_line_is_colnames=False):
        """
        Creates a :ry:factory:`~rayon.data.Dataset` from a file. The
        format of the file is described :ref:`On-disk format
        <data-on-disk-format>`.

        If supplied, *colnames* is a sequence of strings containing
        column names for the data. There must be as many names in the
        sequence as there are columns in the data. If not specified,
        column names from the file will be used; if the file contains
        no column names, the :ry:factory:`~rayon.data.Dataset` will be
        created without column names.

        If supplied, *typemap* is a list of functions that take a
        string representation of an object and return the
        object. There must be as many functions in the sequence as
        there are columns in the data. If not specified, strings
        composed entirely of numbers (optionally with a single period)
        will be converted to numbers, and everything else will be left
        as strings.

        If supplied, *delimiter* is the character used in the file to
        delimit fields in a record. The default delimiter is the pipe
        (``|``) character. The delimiter may also be specified in the
        file format; in that case, this argument is ignored.

        If *first_line_is_colnames* is ``True`` and no other column
        names are supplied in the file (i.e., with a header or a
        specially-commented first line), assume the first line of the
        input is a delimited list of column names. Use this when
        parsing input from external programs (such as the SiLK tools)
        which pass their column names as the first line in a file but
        do not indicate them in any other way.
        """
        return data.Dataset.from_file(
            filename, colnames, typemap, delimiter, first_line_is_colnames)

    new_dataset_from_filename.meta = (data.Dataset, None)

    def new_dataset_from_columns(self, columns, colnames=None,
                                 header=None, sortkey=None):
        """
        Creates a :ry:factory:`~rayon.data.Dataset` from an iterable
        (list, tuple, iterator, etc.) of columns.

        *columns* is an iterable containing columns of data. The
        columns may be lists, tuples, other iterables, or
        :ry:factory:`~rayon.data.Column` objects. All elements in *columns*
        must be of equal length.

        If specified, *colnames* is an iterable of strings. Each item
        represents the name of the corresponding column in the
        dataset. The number of elements in *colnames* should be equal
        to the number of items in *columns*.

        If specified, *header* is a dictionary whose keys and values
        are both strings containing mostly arbitrary metadata about
        the dataset. Some header names, however, are reserved; see
        :ref:`Special Headers <data-special-headers>`.

        If specified, *sortkey* is a function that takes a :ref:`row
        <ref-row-objects>` and returns a comparable. (See the
        :meth:`~rayon.data.Dataset.sort` method for a more precise
        definition.) This function will be used to determine the
        current sort order of the dataset. If not supplied, the
        current sort order will be the order of the items as they
        appear in *columns*.
        """
        return data.Dataset.from_columns(columns, colnames, header, sortkey)

    new_dataset_from_columns.meta = (data.Dataset, None)

    def new_dataset_from_rows(self, rows, colnames=None,
                                 header=None, sortkey=None):
        """
        Creates a :ry:factory:`~rayon.data.Dataset` from an iterable
        (list, tuple, iterator, etc.)  of rows.

        *rows* is an iterable containing rows of data. The rows
        may be lists, tuples, other iterables or dictionaries. All
        elements in *rows* must be of equal length. If *rows*
        contains dictionaries, they must additionally all have the
        same keys.

        If specified, *colnames* is an iterable of strings . Each item
        represents the name of the corresponding column in the
        dataset. The number of elements in colnames should be equal to
        the number of elements in any record in *rows*.

        If *rows* contains dictionaries and *colnames* is supplied,
        every element in *colnames* must be a key in the dictionary
        contents of *rows*. If *colnames* is not supplied, the order
        of columns in the dataset will be the alphanumeric sort order
        of the keys of the dictionaries in *rows*.
        
        If specified, *header* is a dictionary whose keys and values
        are both strings containing mostly arbitrary metadata about
        the dataset. Some header names, however, are reserved; see
        :ref:`Special Headers <data-special-headers>`.

        If specified, *sortkey* is a function that takes a
        :ref:`Row <ref-row-objects>` object and returns a comparable. (See the
        :meth:`sort` method for a more precise definition.) This
        function will be used to determine the current sort order of
        the dataset. If not supplied, the current sort order will be
        insert order.
        """
        
        return data.Dataset.from_rows(rows, colnames, header, sortkey)

    new_dataset_from_rows.meta = (data.Dataset, None)

    def new_column_from_data(self, raw_data):
        """
        Creates a new :ry:factory:`~rayon.data.Column` object from a
        set of data.

        *raw_data* is an iterable (list, tuple, iterator, etc.) containing
        single values of the same data type.
        """
        return data.Column(raw_data)

    new_column_from_data.meta = (data.Column, None)

    def new_column_from_constant(self, constant, length):
        """
        Creates a :ry:factory:`~rayon.data.Column` object from a constant
        value and a length.

        *constant* will be returned for all elements in this column.

        *length* is the length of the column.
        """
        return data.ConstantColumn(constant, length)

    new_column_from_constant.meta = (data.ConstantColumn, None)
    
    #
    # Charts
    #
    
    def new_chart(self, nickname, **kargs):
        """
        Creates a :ref:`chart <ref-charts>` from
        *nickname*. Additional arguments will be passed to the chart subtype
        referred to by *nickname*.

        The *nickname* argument may be one of the following values:

            * :ry:factory:`~rayon.charts.SquareChart`
            * :ry:factory:`~rayon.charts.SimpleTiledChart`
            * :ry:factory:`~rayon.charts.TiledChart`
        
        """
        return charts.create(nickname, self.backend, **kargs)

    new_chart.meta = (charts.Chart, charts.list_chart_types())
    
    @classmethod
    def list_chart_labels(cls):
        """
        Returns ``(label, class)`` tuples for each class that can be
        created with :meth:`Toolbox.new_chart`.
        """
        return charts.list_chart_types()
    
    @classmethod
    def print_chart_labels(cls, stream=sys.stdout):
        """
        Prints to *stream* (``sys.stdout`` by default) all the
        possible values for the *nickname* argument of
        :meth:`Toolbox.new_chart`, and the object types they create.
        """
        print >>stream, "Available chart types:"
        for label, cls in self.list_chart_labels():
            print >>stream, "%15s: rayon.charts.%s" % (label, cls.__name__)
        return

    #
    # Plots
    #

    def new_plot(self, nickname, **kargs):
        """
        Creates a :ref:`plot <ref-plots>` from *nickname*. Additional
        arguments will be passed to the plot subtype referred to by
        *nickname*.

        The *nickname* argument may be one of the following values:

            * :ry:factory:`~rayon.plots.VerticalBarPlot`
            * :ry:factory:`~rayon.plots.FieldPlot`
            * :ry:factory:`~rayon.plots.FilledLinePlot`
            * :ry:factory:`~rayon.plots.HorizontalBarPlot`
            * :ry:factory:`~rayon.plots.LabeledScatterPlot`
            * :ry:factory:`~rayon.plots.LinePlot`
            * :ry:factory:`~rayon.plots.ScatterPlot`
        
        """
        return plots.create(nickname, **kargs)

    # What's a DocumentedPlot? See rayon/plots.py
    new_plot.meta = (plots.DocumentedPlot, plots.list_plot_types())

    @classmethod
    def list_plot_labels(cls):
        """
        Returns ``(label, class)`` tuples for each class that can be
        created with :meth:`Toolbox.new_plot`.
        """
        return plots.list_plot_types()

    @classmethod
    def print_plot_labels(cls, stream=sys.stdout):
        """
        Prints to *stream* (``sys.stdout`` by default) all the
        possible values for the *nickname* argument of
        :meth:`Toolbox.new_plot`, and the object types they create.
        """
        print >>stream, "Available plot types:"
        for label, cls in self.list_plot_labels():
            print >>stream, "%15s: rayon.plots.%s" % (label, cls.__name__)
        return
    

    #
    # Scales
    #
    
    def new_scale(self, nickname, **kargs):
        """
        Creates a :ref:`scale <ref-scales>` from
        *nickname*. Additional arguments will be passed to the scale
        subtype referred to by *nickname*.

        The *nickname* argument may be one of the following values:

            * :ry:factory:`~rayon.scales.CategoricalScale`
            * :ry:factory:`~rayon.scales.CategoricalRangeScale`
            * :ry:factory:`~rayon.scales.CountingLogScale`
            * :ry:factory:`~rayon.scales.CountingLogGradientScale`
            * :ry:factory:`~rayon.scales.LinearGradientScale`
            * :ry:factory:`~rayon.scales.LinearScale`
            * :ry:factory:`~rayon.scales.LinearBinnedScale`
            * :ry:factory:`~rayon.scales.LogScale`
            * :ry:factory:`~rayon.scales.LogGradientScale`
            * :ry:factory:`~rayon.scales.TimeBinnedScale`
        
        """
        return scales.create(nickname, **kargs)

    new_scale.meta = (scales.Scale, scales.list_scale_types())
    
    @classmethod
    def list_scale_labels(cls):
        """
        Returns ``(label, class)`` tuples for each class that can be
        created with :meth:`Toolbox.new_scale`.
        """
        return scales.list_scale_types()
    
    @classmethod
    def print_scale_labels(cls, stream=sys.stdout):
        """
        Prints to *stream* (``sys.stdout`` by default) all the
        possible values for the *nickname* argument of
        :meth:`Toolbox.new_scale`, and the object types they create.
        """
        print >>stream, "Available scale types:"
        for label, cls in self.list_scale_labels():
            print >>stream, "%15s: rayon.scales.%s" % (label, cls.__name__)
        return


    #
    # Borders
    #
    
    def new_border(self, nickname, **kargs):
        """
        Creates a :ref:`border <ref-borders>` from
        *nickname*. Additional arguments will be passed to the border
        subtype referred to by *nickname*.

        The *nickname* argument may be one of the following values:

            * :ry:factory:`~rayon.borders.HorizontalLabeledBorder`
            * :ry:factory:`~rayon.borders.HorizontalLineBorder`
            * :ry:factory:`~rayon.borders.HorizontalBorder`
            * :ry:factory:`~rayon.borders.HorizontalSplitBorder`
            * :ry:factory:`~rayon.borders.NoBorder`
            * :ry:factory:`~rayon.borders.VerticalBorder`
            * :ry:factory:`~rayon.borders.VerticalLabeledBorder`
            * :ry:factory:`~rayon.borders.VerticalLineBorder`
            * :ry:factory:`~rayon.borders.VerticalSplitBorder`
        """
        return borders.create(nickname, **kargs)

    new_border.meta = (borders.Border, borders.list_border_types())
    
    
    @classmethod
    def list_border_labels(cls):
        """
        Returns ``(label, class)`` tuples for each class that can be
        created with :meth:`Toolbox.new_border`.
        """
        return borders.list_border_types()
    
    
    @classmethod
    def print_border_labels(cls, stream=sys.stdout):
        """
        Prints to *stream* (``sys.stdout`` by default) all the
        possible values for the *nickname* argument of
        :meth:`Toolbox.new_border`, and the object types they create.
        """
        print >>stream, "Available border types:"
        for label, cls in self.list_border_labels():
            print >>stream, "%15s: rayon.borders.%s" % (label, cls.__name__)
        return

    #
    # Lines
    #
    
    def new_line(self, nickname, **kargs):
        """
        Creates Rayon line from *nickname*.

        The *nickname* argument may be one of the following values:

            * :ry:factory:`~rayon.lines.Dashed`
            * :ry:factory:`~rayon.lines.DotDash`
            * :ry:factory:`~rayon.lines.Dotted`
            * :ry:factory:`~rayon.lines.Null`
            * :ry:factory:`~rayon.lines.Solid`


        *color* is the color of the line.

        *width* is the width of the line in pixels/points.

        If *fixup* is ``True``, Rayon will attempt to "fix up"
        perfectly horizontal and vertical lines to avoid
        anti-aliasing. This behavior is disabled if *fixup* is
        ``False``.  (See :ref:`ref-lines-fixup`.)
        """
        return lines.create(nickname, **kargs)

    new_line.meta = (lines.Line, lines.list_line_types())
    
    @classmethod
    def list_line_labels(cls):
        """
        Returns ``(label, class)`` tuples for each class that can be
        created with :meth:`Toolbox.new_line`.
        """
        return lines.list_line_types()
    
    @classmethod
    def print_line_labels(cls, stream=sys.stdout):
        """
        Prints to *stream* (``sys.stdout`` by default) all the
        possible values for the *nickname* argument of
        :meth:`Toolbox.new_line`, and the object types they create.
        """
        print >>stream, "Available line types:"
        for label, cls in self.list_line_labels():
            print >>stream, "%15s: rayon.lines.%s" % (label, cls.__name__)
        return

    def new_line_from_spec(self, spec):
        """
        Creates a Rayon line from *spec*, as described in
        :ref:`specifying-lines`.

        *spec* is a string line specification.

        If *fixup* is ``True``, Rayon will attempt to "fix up"
        perfectly horizontal and vertical lines to avoid
        anti-aliasing. This behavior is disabled if *fixup* is
        ``False``. (See :ref:`ref-lines-fixup`.)
        """
        return lines.create_from_spec(spec)

    new_line_from_spec.meta = (lines.Line, None)
    
    #
    # Markers
    #
    
    def new_marker(self, nickname, **kargs):
        """
        Creates a :ref:`marker <ref-markers>` from
        *nickname*. Additional arguments will be passed to the marker
        subtype referred to by *nickname*.
        
        The *nickname* argument may be one of the following values:

            * :ry:factory:`~rayon.markers.Arrow`
            * :ry:factory:`~rayon.markers.Circle`
            * :ry:factory:`~rayon.markers.Cross`
            * :ry:factory:`~rayon.markers.Dot`
            * :ry:factory:`~rayon.markers.HorizontalLine`
            * :ry:factory:`~rayon.markers.NullMarker`
            * :ry:factory:`~rayon.markers.VerticalLine`
        """
        return markers.create(nickname, **kargs)

    new_marker.meta = (markers.Marker, markers.list_marker_types())
    
    @classmethod
    def list_marker_labels(cls):
        """
        Returns ``(label, class)`` tuples for each class that can be
        created with :meth:`Toolbox.new_marker`.
        """
        return markers.list_marker_types()
    
    @classmethod
    def print_marker_labels(cls, stream=sys.stdout):
        """
        Prints to *stream* (``sys.stdout`` by default) all the
        possible values for the *nickname* argument of
        :meth:`Toolbox.new_marker`, and the object types they create.
        """
        print >>stream, "Available marker types:"
        for label, cls in self.list_marker_labels():
            print >>stream, "%15s: rayon.markers.%s" % (label, cls.__name__)
        return


    #
    # Gridlines
    #
    
    def new_gridlines(self, nickname, **kargs):
        """
        Creates a gridlines from *nickname*. The *nickname* argument
        may be one of the following values:

            * :ry:factory:`~rayon.grids.HorizontalGridLines`
            * :ry:factory:`~rayon.grids.VerticalGridLines`

        Additional arguments beyond *nickname* are scale or data
        specifications as they would be given to plots. (See
        :ref:`ref-gridlines-creating` and :ref:`ref-plots-creating`.)
            
        """
        return grids.create(nickname, **kargs)

    new_gridlines.meta = (grids.GridLines, grids.list_gridlines_types())

    def new_gridlines_from_border(self, nickname,
                                  border, tickset_idx=0, **kargs):
        """
        Creates grid lines from *nickname*. The ticksets in *border*
        will be used to generate the positions for the lines.

        The *nickname* argument may be one of the following values:

            * :ry:factory:`~rayon.grids.HorizontalGridLines`
            * :ry:factory:`~rayon.grids.VerticalGridLines`

        *border* is a :ref:`tickable border <ref-borders>` containing
        at least one tick set.

        If supplied, *tickset_idx* is the index of the tickset to use
        to generate grid lines in *border*.

        Additional arguments are scale or data specifications as they
        would be given to plots. (See :ref:`ref-gridlines-creating`
        and :ref:`ref-plots-creating`.)
        """
        return grids.create_from_border(nickname,
                                        border, tickset_idx, **kargs)

    #new_gridlines_from_border.meta = (grids.GridLines,
    #                                  grids.list_gridlines_types())
    
    def new_gridlines_from_tick_tuples(self, nickname,
                                       scale, tick_tuples, **kargs):
        """
        Creates grid lines from *nickname*. The tick positions in
        *tick_tuples* will be used to generate the positions for the
        lines

        The *nickname* argument may be one of the following values:

            * :class:`~rayon.grids.HorizontalGridLines`
            * :class:`~rayon.grids.VerticalGridLines`

        *tick_tuples* will be used to generate the data for the
        ``pos`` axis of the grid lines.

        *scale* will be used as the scale for the ``pos`` axis of the
        grid lines.
        
        Additional arguments are scale or data specifications as they
        would be given to plots. (See :ref:`ref-gridlines-creating`
        and :ref:`ref-plots-creating`.)
        """
        return grids.create_from_tick_tuples(nickname,
                                             scale, tick_tuples, **kargs)

    #new_gridlines_from_tick_tuples.meta = (grids.GridLines,
    #                                       grids.list_gridlines_types())
    
    @classmethod
    def list_gridlines_labels(cls):
        """
        Returns ``(label, class)`` tuples for each class that can be
        created with :meth:`Toolbox.new_gridlines`,
        :meth:`Toolbox.new_gridlines_from_border` and
        :meth:`Toolbox.new_gridlines_from_tick_tuples`.
        """
        return grids.list_gridlines_types()

    @classmethod
    def print_gridlines_labels(cls, stream=sys.stdout):
        """
        Prints to *stream* (``sys.stdout`` by default) all the
        possible values for the *nickname* argument of
        :meth:`Toolbox.new_gridlines`,
        :meth:`Toolbox.new_gridlines_from_border` and
        :meth:`Toolbox.new_gridlines_from_tick_tuples`; and the object
        types they create.
        """
        print >>stream, "Available gridline types:"
        for label, cls in self.list_gridlines_labels():
            print >>stream, "%15s: rayon.grids.%s" % (label, cls.__name__)
        return


    #
    # Tickmarkers
    #
    
    def new_tickmarker(self, nickname, **kargs):
        # Deprecated, as are TickMarkers. Use LabeledMarkers instead.
        return tickmarkers.create(nickname, **kargs)

    new_tickmarker.meta = (tickmarkers.TickMarker,
                           tickmarkers.list_tickmarker_types())
    
    @classmethod
    def list_tickmarker_labels(cls, return_simple=True,
                               return_nonsimple=False):
        """
        Returns ``(label, class)`` tuples for each class that can be
        created with :meth:`Toolbox.new_tickmarker`.

        If *return_simple* is ``True``, this method will return the
        objects implementing the "simple" interface for object
        creation (i.e.,
        :class:`~rayon.tickmarkers.SimpleHorizontalTickMarker` and
        :class:`~rayon.tickmarkers.SimpleVerticalTickMarker`). If
        *return_nonsimple* is ``True``, this method will return the
        objects implementing the "non-simple" interface for object
        creation (i.e., :`~rayon.tickmarkers.HorizontalTickMarker` and
        :class:`~rayon.tickmarkers.VerticalTickMarker`).
        """
        return tickmarkers.list_tickmarker_types()

    
    @classmethod
    def print_tickmarker_labels(cls, stream=sys.stdout):
        """
        Prints to *stream* (``sys.stdout`` by default) all the
        possible values for the *nickname* argument of
        :meth:`Toolbox.new_tickmarker`.
        """
        print >>stream, "Available gridline types:"
        for label, cls in self.list_gridlines_labels():
            print >>stream, "%15s: rayon.grids.%s" % (label, cls.__name__)
        return
    

    #
    # Ticksets
    #

    def new_tickset(self, nickname, data=None, spec=None, formatter=None,
                    **kargs):
        """
        Creates a :class:`~rayon.ticks.TickSet` object from
        *nickname*. Additional arguments will be passed to the tickset
        subtype referred to by *nickname*.

        The *nickname* argument may be one of the following values:

           * :ry:factory:`~rayon.ticks.TickSet`"
           * :ry:factory:`~rayon.ticks.RangeTickSet`"
        """
        # This method is deprecated. Use new_tickset_from_tuples or
        # new_tickset from_spec instead.
        # TODO: remove for Rayon 2.0
        return ticks.create(nickname, data, spec, formatter, **kargs)

    #new_tickset.meta = (ticks.TickSetBase, ticks.list_tickset_types())

    def new_tickset_from_tuples(self, nickname, tick_tuples, scale,
                                labeledmarker, alignment=.5):
        """
        Creates a tick set from a set of tick tuples representing
        where and how ticks should be created.

        The *nickname* argument may be one of the following values:

            * :ry:factory:`~rayon.ticks.TickSet`
            * :ry:factory:`~rayon.ticks.RangeTickSet`

        *tick_tuples* is an iterable of 2-tuples ``(value, label)``.
        ``value`` should be a valid input to *scale*. ``label`` is a
        string representation of ``value``. If ``label`` is ``None``,
        Rayon will convert ``value`` to a string using :class:`str`
        and use that as the label.

        *scale* is a Rayon :ref:`scale <ref-scales>` or a function
        that can serve as a scale. The output of *scale* should be a
        number between 0 and 1 if *nickname* is ``default``. If
        *nickname* is ``ranged``, *scale* should output a tuple
        ``(min, max)`` describing the minimum and maximum bounds of a
        range between 0 and 1. (See :ref:`ref-scales-range-scales`)

        *labeledmarker* is a :ref:`labeled marker
        <ref-labeledmarkers>` that will be used to draw the tick mark.

        If *nickname* is ``ranged`` and if supplied, *alignment* is a
        number between 0 and 1 specifying the alignment of the tick
        mark within the allotted range. A value of 0 for *alignment*
        will put the tick mark at the leftmost point in the range; a
        value of 1 will put it at the rightmost point.
        """
        tick_tuples = tuple(t for t in tick_tuples)
        return ticks.create_from_tuples(nickname, tick_tuples=tick_tuples,
                            scale=scale, marker=labeledmarker,
                                        alignment=alignment)

    new_tickset_from_tuples.meta = (ticks.TickSetBase,
                                    ticks.list_tickset_types())

    def new_tickset_from_spec(self, nickname, tick_spec,
                              col, scale, labeledmarker, formatter=None,
                              alignment=.5):
        """
        Creates a tick set from a string specification and a set of
        sortable data.

        The *nickname* argument may be one of the following values:

            * :ry:factory:`~rayon.ticks.TickSet`
            * :ry:factory:`~rayon.ticks.RangeTickSet`


        *tick_spec* is a string of tick specifications, separated by
        commas. A tick specification may be one of the following:

        .. csv-table::
            :header: "Spec", "Definition"

            "``at(N,M)``","The literal value *N*, *M*, etc.. ('Put a tick at *N*')"
            "``every(N)``","The first point in the data, and every subsequent point at interval N ('Put a tick every *N* (e.g., every 10 minutes)')"
            "``ge(N)``","All points in the data greater than or equal to *N*"
            "``gt(N)``","All points in the data greater than *N*"
            "``le(N)``","All points in the data less than or equal to *N*"
            "``lt(N)``","All points in the data less than *N*"
            "``max``","The largest value in the data"
            "``med``","The median value in the data (a synonym for ``p(50)``)"
            "``min``","The smallest value in the data"
            "``n(N)``","Place *N* ticks evenly along the scale, at 'nice' places"
            "``p(N)``","The value for percentile *N*"
            "``smax``","The largest value on the scale"
            "``smin``","The smallest value on the scale"

        *col* is a column of data. (Either an iterable or a
        :ry:factory:`~rayon.data.Column` object.) This data will be
        used with the specifications to determine the tick positions.

        *scale* is the scale to use to place the tickmark. (See
        :ref:`ref-scales`.) It will also be used, if necessary, to
        compute the ``smax`` and ``smin`` tick specifications.The
        output of *scale* should be a number between 0 and 1 if
        *nickname* is ``default``. If *nickname* is ``ranged``,
        *scale* should output a tuple ``(min, max)`` describing the
        minimum and maximum bounds of a range between 0 and 1. (See
        :ref:`ref-scales-range-scales`).

        *labeledmarker* is a
        :ref:`labeled marker <ref-labeledmarkers>` that will be
        used to draw the tick marks.

        If supplied, *formatter* is either a printf-style format string or a
        function that takes a value in *col* and returns a string that
        will be used to label the tick mark with that value. If
        *formatter* is ``None``, Rayon supply a formatter that
        converts the input value to a string using :class:`str` and
        use that as the label.

        If *nickname* is ``ranged`` and if supplied, *alignment* is a
        number between 0 and 1 specifying the alignment of the tick
        mark within the allotted range. A value of 0 for *alignment*
        will put the tick mark at the leftmost point in the range; a
        value of 1 will put it at the rightmost point.
        """
        return ticks.create_from_new_spec(
            tickset_type=nickname, col=col, spec=tick_spec, formatter=formatter,
            scale=scale, marker=labeledmarker, alignment=alignment)

    new_tickset_from_spec.meta = (ticks.TickSetBase,
                                  ticks.list_tickset_types())
    
    @classmethod
    def list_tickset_labels(cls):
        """
        Returns ``(label, class)`` tuples for each class that can be
        created with :meth:`Toolbox.new_tickset`.
        """
        return ticks.list_tickset_types()
    
    @classmethod
    def print_tickset_labels(cls, stream=sys.stdout):
        """
        Prints to *stream* (``sys.stdout`` by default) all the
        possible values for the *nickname* argument of
        :meth:`Toolbox.new_tickset`, and the object types they create.
        """
        print >>stream, "Available tickset types:"
        for label, cls in self.list_tickset_labels():
            print >>stream, "%15s: rayon.ticks.%s" % (label, cls.__name__)
        return


    #
    # Labelers
    #

    def new_labeler(self, txt=None, **kargs):
        """
        Creates a new :ref:`labeler <ref-text>`.

        If *txt* is supplied, the labeler will be static -- it will
        only ever render *txt*. Static labelers are convenient to use
        for simple labels that will be drawn only once, such as chart
        titles and captions. If *txt* is not supplied, the labeler can
        be drawn many times with different text. In this case, the
        labeler is used as a template of text properties.

        The remaining arguments are described in detail in
        :ref:`ref-text-specifying-text-properties`.
        """
        if txt is not None:
            return text.StaticText(txt, **kargs)
        else:
            return text.Labeler(**kargs)

    new_labeler.meta = (text.Labeler, None)

    def new_label(self, txt=None, **kargs):
        # Deprecated. Use new_labeler instead
        # TODO: remove for Rayon 2.0
        return self.new_labeler(txt, **kargs)

    #new_label.meta = (text.Labeler, None)
    
        
    def new_labeled_marker(self,
                          marker_type="dot", marker_args={},
                          labeler_args={}, **kargs):
        """
        Creates a new :ref:`labeled marker <ref-labeledmarkers>`.

        *marker* is a :ref:`marker <ref-markers>` that will be used to
        generate the marked part of the labeled marker.

        *labeler* is a :ref:`labeler <ref-labelers>` that will be used
        to generate the labeled part of the labeled marker.

        If supplied, *label_position* describes the position of the
        label relative to the marker. Valid values are ``n`` (above),
        ``s`` (below), ``e`` (to the left) and ``w`` (to the right).

        If supplied, *label_spacing* is the amount of space between
        the label and the marker, in device units (points or pixels).
        """
        # NOTE: marker_type and marker_args are no longer documented.
        # TODO: remove marker_type and marker_args for Rayon 2.0
        if not kargs.has_key('marker'):
            kargs['marker'] = self.new_marker(marker_type, **marker_args)
        if not kargs.has_key('labeler'):
            kargs['labeler'] = self.new_label(**labeler_args)

        if kargs['marker'] is None:
            kargs['marker'] = self.new_marker('none')
        return markers.LabeledMarker(**kargs)

    new_labeled_marker.meta = (markers.LabeledMarker, None)

    # Non-object-creation methods

    def parse_color(self, color_spec):
        """
        If *color_spec* is a string in specifying a color, converts it
        into into a tuple of 4 ints 0-255, representing a color in
        RGBA format. If *color_spec* is already a tuple, returns it.

        The syntax of the color specification is identical to that
        specified by
        `<http://www.w3.org/TR/CSS2/syndata.html#color-units>` and
        '<http://www.w3.org/TR/css3-color>', with the exception that
        alpha channels are supported in hexadecimal notation as well
        as functional notation.
        
        As an example, here are all the valid ways to specify full red
        (100% red, 0% green, 0% blue, 100% opaque):
        
           * ``red``
           * ``#f00``
           * ``#f00f``
           * ``#ff0000``
           * ``#ff0000ff``
           * ``#F00``
           * ``#F00F``
           * ``#FF0000``
           * ``#FF0000FF``
           * ``rgb(255, 0, 0)``
           * ``rgba(255, 0, 0, 1)``
           * ``rgb(100%, 0%, 0%)``
           * ``rgba(100%, 0%, 0%, 1)``
        """
        return miscutils.parse_color(color_spec)

    # Metainfo interface (undocumented, used by documentation)

    ctor_fns = None
    @classmethod
    def _get_metainfo(cls):
        if cls.ctor_fns is None:
            # A list of (fn, returntype) pairs
            ctor_fns = [(getattr(cls, name), getattr(cls, name).meta)
                        for name in dir(cls)
                        if hasattr(getattr(cls, name), "meta")]
        return ctor_fns


    @classmethod
    def can_document(cls, clsobj):
        # Roll through functions on cls that have .meta and return
        # True if one or more of them return objects of which clsobj
        # is a subclass
        for fn, (returntype, nickmap) in cls._get_metainfo():
            if issubclass(clsobj, returntype):
                return True
        return False

    @classmethod
    def resolve_class(cls, clsobj):
        # Return a list of (function, nickname) pairs corresponding to
        # the functions and nicknames that will return an object of
        # type *clsobj*

        rc = []
        for fn, (returntype, nickmap) in cls._get_metainfo():
            if issubclass(clsobj, returntype):
                if nickmap is None:
                    nicks = tuple()
                else:
                    # Compile the nicknames that return objects with the
                    # same type as this one.
                    nicks = tuple(nick for nick, ret_obj in nickmap
                                  if ret_obj == clsobj)
                rc.append((fn, nicks))
        return rc

class MiniToolbox(Toolbox):
    """
    A reduced-functionality version of the :class:`Toolbox` that
    doesn't know or care about the drawing backend. All methods
    related to rendering to the backend are disabled and will raise
    ``NotImplemented``. This object can be used in places where the
    backend is unknown, such as in the :meth:`~plots.Plot.draw_`
    method of the :class:`~plots.Plot` object.

    All the same constructor classmethods will return a
    :class:`MiniToolbox` object successfully, though any backend
    information will be ignored. It is preferred, however, to use the
    class's ``__init__`` method.
    """
    @classmethod
    def for_backend(cls, backend_name): return cls()

    @classmethod
    def for_file(cls): return cls()

    @classmethod
    def for_gui(cls): return cls()

    def __init__(self): Toolbox.__init__(self, None)

    def new_page_from_buffer(self, buff):
        "Raises exception. Disabled in this class."
        raise NotImplemented(
            "Method not implemented in MiniToolbox. "
            "Use Toolbox for methods that use the rendering backend.")

    def new_page_from_filename(self, filename, width, height):
        "Raises exception. Disabled in this class."
        raise NotImplemented(
            "Method not implemented in MiniToolbox. "
            "Use Toolbox for methods that use the rendering backend.")

    def new_chart(self, nickname, **kargs):
        "Raises exception. Disabled in this class."
        raise NotImplemented(
            "Method not implemented in MiniToolbox. "
            "Use Toolbox for methods that use the rendering backend.")
        
    

__all__ = ['Toolbox', 'MiniToolbox']
